from __future__ import annotations

import warnings
from pathlib import Path
from typing import Callable, Dict, Optional, Tuple

from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.fs.path import path_absolute
from pharmpy.internals.immutable import frozenmapping
from pharmpy.internals.math import triangular_root
from pharmpy.model import (
    Assignment,
    ColumnInfo,
    DataInfo,
    DatasetError,
    EstimationStep,
    EstimationSteps,
    JointNormalDistribution,
    ModelSyntaxError,
    NormalDistribution,
    Parameter,
    Parameters,
    RandomVariables,
    Statements,
)

from .advan import _compartmental_model, des_assign_statements
from .dataset import read_nonmem_dataset
from .nmtran_parser import NMTranControlStream
from .table import NONMEMTableFile, PhiTable


def parse_thetas(control_stream):
    names = []
    bounds = []
    inits = []
    fixs = []
    for theta_record in control_stream.get_records('THETA'):
        bounds.extend(theta_record.bounds)
        inits.extend(theta_record.inits)
        fixs.extend(theta_record.fixs)
        names.extend(theta_record.comment_names)
    fixs = _fix_thetas_with_same_bounds(bounds, inits, fixs)
    return names, bounds, inits, fixs


def _fix_thetas_with_same_bounds(bounds, inits, fixs):
    # NONMEM will autofix thetas having all bounds equal
    new_fix = []
    for bound, init, fix in zip(bounds, inits, fixs):
        lower = bound[0]
        upper = bound[1]
        if lower == upper == init:
            new_fix.append(True)
        else:
            new_fix.append(fix)
    return new_fix


def parse_omegas_sigmas(control_stream, record_name):
    blocks = []
    for record in control_stream.get_records(record_name):
        curblocks = record.parse()
        blocks.extend(curblocks)
    return blocks


def parameters_from_blocks(blocks, all_names, record_name):
    row = 1
    col = 1
    prev_size = None
    parameters = []
    name_map = {}
    for names, inits, fix, same in blocks:
        if same:
            if prev_size is None:
                raise ModelSyntaxError(f"First {record_name} block cannot be SAME")
            row += prev_size
            col += prev_size
        else:
            block_row = row
            col = row
            for i, name in enumerate(names):
                if name in all_names:
                    duplicated_name = name
                    name = None
                else:
                    duplicated_name = None
                if name is None:
                    name = f'{record_name}_{row}_{col}'
                    while name in all_names:
                        name += "_"
                if duplicated_name is not None:
                    warnings.warn(
                        f'The parameter name {duplicated_name} is duplicated. '
                        f'Falling back to using {name} instead.'
                    )
                all_names.add(name)
                nonmem_name = f'{record_name}({row},{col})'
                name_map[nonmem_name] = name
                lower = 0 if row == col else None
                parameter = Parameter.create(name, init=inits[i], lower=lower, upper=None, fix=fix)
                parameters.append(parameter)
                if row == col:
                    row += 1
                    col = block_row
                else:
                    col += 1
            prev_size = row - block_row
    return parameters, name_map


def rvs_from_blocks(abbr_names, blocks, parameters, rvtype):
    next_same = False
    parameters_index = 0
    eta_index = 1
    rvs = []
    previous_cov = None
    name_map = {}
    for block_index, (_, inits, _, same) in enumerate(blocks):
        try:
            next_block = blocks[block_index + 1]
        except IndexError:
            next_same = False
        else:
            next_same = next_block[3]

        if not same:
            n = triangular_root(len(inits))

        if rvtype == 'EPS':
            level = 'ruv'
        else:
            if same or next_same:
                level = 'iov'
            else:
                level = 'iiv'

        names = []
        for i in range(n):
            nonmem_name = f'{rvtype}({eta_index + i})'
            if nonmem_name in abbr_names:
                name = abbr_names[nonmem_name]
            else:
                name = f'{rvtype}_{eta_index + i}'
            name_map[nonmem_name] = name
            if rvtype == 'EPS':
                alt_name = f'ERR({eta_index + i})'
                name_map[alt_name] = name
            names.append(name)

        if same:
            cov = previous_cov
        else:
            if n == 1:
                cov = parameters[parameters_index].symbol
                parameters_index += 1
            else:
                cov = sympy.zeros(n)
                for row in range(n):
                    for col in range(0, row + 1):
                        symb = parameters[parameters_index].symbol
                        parameters_index += 1
                        cov[row, col] = symb
                        cov[col, row] = symb

        if n == 1:
            dist = NormalDistribution.create(names[0], level, 0, cov)
        else:
            means = [0] * n
            dist = JointNormalDistribution.create(names, level, means, cov)

        rvs.append(dist)
        previous_cov = cov

        eta_index += n

    return RandomVariables.create(rvs), name_map


def parse_parameters(control_stream, statements):
    symbols = statements.free_symbols
    all_names = {s.name for s in symbols}
    theta_names, theta_bounds, theta_inits, theta_fixs = parse_thetas(control_stream)
    abbr_map = control_stream.abbreviated.translate_to_pharmpy_names()
    theta_parameters = []
    name_map = {}
    for i, name in enumerate(theta_names):
        nonmem_name = f'THETA({i + 1})'
        if nonmem_name in abbr_map:
            name = abbr_map[nonmem_name]
        if name in all_names:
            duplicated_name = name
            name = None
        else:
            duplicated_name = None
        if name is None:
            name = f'THETA_{i + 1}'
            while name in all_names:
                name += "_"
        if duplicated_name is not None:
            warnings.warn(
                f'The parameter name {duplicated_name} is duplicated. '
                f'Falling back to using {name} instead.'
            )
        name_map[nonmem_name] = name
        all_names.add(name)
        parameter = Parameter.create(
            name,
            init=theta_inits[i],
            lower=theta_bounds[i][0],
            upper=theta_bounds[i][1],
            fix=theta_fixs[i],
        )
        theta_parameters.append(parameter)

    omega_blocks = parse_omegas_sigmas(control_stream, "OMEGA")
    sigma_blocks = parse_omegas_sigmas(control_stream, "SIGMA")
    omega_parameters, omega_map = parameters_from_blocks(omega_blocks, all_names, 'OMEGA')
    sigma_parameters, sigma_map = parameters_from_blocks(sigma_blocks, all_names, 'SIGMA')
    name_map.update(omega_map)
    name_map.update(sigma_map)
    parameters = Parameters.create(theta_parameters + omega_parameters + sigma_parameters)

    etas, eta_map = rvs_from_blocks(abbr_map, omega_blocks, omega_parameters, 'ETA')
    epsilons, eps_map = rvs_from_blocks(abbr_map, sigma_blocks, sigma_parameters, 'EPS')
    name_map.update(eta_map)
    name_map.update(eps_map)

    rvs = etas + epsilons

    return parameters, rvs, name_map


def parse_statements(
    di: DataInfo,
    dataset: Callable[[], pd.DataFrame],
    control_stream: NMTranControlStream,
) -> Tuple[Statements, Optional[Dict[str, int]]]:
    rec = control_stream.get_pred_pk_record()
    statements = rec.statements

    des = control_stream.get_des_record()
    error = control_stream.get_error_record()
    comp_map = None

    if error:
        sub = control_stream.get_records('SUBROUTINES')[0]
        des_assign = des_assign_statements(control_stream, des)
        if des_assign is not None:
            for s in des_assign:
                statements += Assignment(s.lhs, s.rhs)
        comp = _compartmental_model(di, dataset, control_stream, sub.advan, sub.trans, des)
        trans_amounts = {}
        if comp is None:
            raise ValueError("Error in parsing ODE System")
        else:
            cm, link, comp_map = comp
            statements += [cm, link]
            if des:
                rec_model = control_stream.get_records('MODEL')[0]
                comps = [c for c, _ in rec_model.compartments()]
                for i, c in enumerate(comps, 1):
                    trans_amounts[sympy.Symbol(f"A({i})")] = sympy.Function(f'A_{c}')(
                        sympy.Symbol('t')
                    )
                    trans_amounts[sympy.Symbol(f"A_0({i})")] = sympy.Function(f'A_{c}')(0)
            else:
                for i, amount in enumerate(cm.amounts, start=1):
                    trans_amounts[sympy.Symbol(f"A({i})")] = amount
                    trans_amounts[sympy.Symbol(f"A_0({i})")] = sympy.Function(amount.name)(0)

        statements += error.statements
        if trans_amounts:
            statements = statements.subs(trans_amounts)

    return statements, comp_map


def convert_dvs(statements, control_stream):
    # Conversion of IF (DVID.EQ.n) THEN to non-piecewise
    # could partly be done in code_record
    after = statements.error
    kept = []
    dvs = frozenmapping({sympy.Symbol('Y'): 1})
    obs_trans = frozenmapping({sympy.Symbol('Y'): sympy.Symbol('Y')})
    change = False
    yind = after.find_assignment_index('Y')
    if yind is None:
        return statements, dvs, obs_trans
    yind = yind - 1
    for s in after:
        expr = s.expression
        if isinstance(expr, sympy.Piecewise) and sympy.Symbol("DVID") in expr.free_symbols:
            cond = expr.args[0][1]
            if cond.lhs == sympy.Symbol("DVID") and cond.rhs == 1:
                ass1 = s.replace(symbol=sympy.Symbol('Y_1'), expression=expr.args[0][0])
                ass2 = s.replace(symbol=sympy.Symbol('Y_2'), expression=expr.args[1][0])
                if expr.args[0][0] not in statements.free_symbols:
                    kept.append(ass1)
                    dv_1 = 'Y_1'
                else:
                    dv_1 = str(expr.args[0][0])
                if expr.args[1][0] not in statements.free_symbols:
                    kept.append(ass2)
                    dv_2 = 'Y_2'
                else:
                    dv_2 = str(expr.args[1][0])
                dvs = frozenmapping({sympy.Symbol(dv_1): 1, sympy.Symbol(dv_2): 2})
                obs_trans = frozenmapping({dv: dv for dv in dvs.keys()})
                change = True
                continue
        kept.append(s)
    after = Statements(tuple(kept))
    if statements.ode_system is not None:
        statements = statements.before_odes + statements.ode_system + after
        if change:  # $ERROR
            rec = control_stream.get_records('ERROR')[0]
    else:
        statements = after
        if change:  # $PRED
            rec = control_stream.get_records('PRED')[0]
    if change:
        rec._statements = after[1:]
        first = True
        inds = []
        for ni, nj, si, sj in rec._index:
            if first and si == yind:
                first = False
                ind = (ni, nj, si, si + 2)
            else:
                if first:
                    ind = (ni, nj, si, si)
                else:
                    ind = (ni, nj, si + 2, sj + 2)
            inds.append(ind)
        rec._index = inds
    return statements, dvs, obs_trans


def parse_value_type(control_stream, statements):
    ests = control_stream.get_records('ESTIMATION')
    # Assuming that a model cannot be fully likelihood or fully prediction
    # at the same time
    for est in ests:
        if est.likelihood:
            tp = 'LIKELIHOOD'
            break
        elif est.loglikelihood:
            tp = '-2LL'
            break
    else:
        tp = 'PREDICTION'
    f_flag = sympy.Symbol('F_FLAG')
    if f_flag in statements.free_symbols:
        tp = f_flag
    return tp


def parse_description(control_stream) -> str:
    rec = control_stream.get_records('PROBLEM')[0]
    return rec.title


def parse_estimation_steps(control_stream, random_variables) -> EstimationSteps:
    steps = []
    records = control_stream.get_records('ESTIMATION')
    covrec = control_stream.get_records('COVARIANCE')
    solver, tol, atol = parse_solver(control_stream)

    # Read eta and epsilon derivatives
    etaderiv_names = None
    epsilonderivs_names = None
    table_records = control_stream.get_records('TABLE')
    for table in table_records:
        etaderivs = table.eta_derivatives
        if etaderivs:
            etas = random_variables.etas
            etaderiv_names = [etas.names[i - 1] for i in etaderivs]
        epsderivs = table.epsilon_derivatives
        if epsderivs:
            epsilons = random_variables.epsilons
            epsilonderivs_names = [epsilons.names[i - 1] for i in epsderivs]

    for record in records:
        value = record.get_option('METHOD')
        if value is None or value == '0' or value == 'ZERO':
            name = 'fo'
        elif value == '1' or value == 'CONDITIONAL' or value == 'COND':
            name = 'foce'
        else:
            name = value
        interaction = False
        evaluation = False
        maximum_evaluations = None
        parameter_uncertainty_method = None
        laplace = False
        isample = None
        niter = None
        auto = None
        keep_every_nth_iter = None

        if record.has_option('INTERACTION') or record.has_option('INTER'):
            interaction = True
        maxeval_opt = record.get_option('MAXEVAL') if not None else record.get_option('MAXEVALS')
        if maxeval_opt is not None:
            if (name.upper() == 'FO' or name.upper() == 'FOCE') and int(maxeval_opt) == 0:
                evaluation = True
            else:
                maximum_evaluations = int(maxeval_opt)
        eval_opt = record.get_option('EONLY')
        if eval_opt is not None and int(eval_opt) == 1:
            evaluation = True
        if covrec:
            parameter_uncertainty_method = 'SANDWICH'
        if record.has_option('LAPLACIAN') or record.has_option('LAPLACE'):
            laplace = True
        if record.has_option('ISAMPLE'):
            isample = int(record.get_option('ISAMPLE'))
        if record.has_option('NITER'):
            niter = int(record.get_option('NITER'))
        if record.has_option('AUTO'):
            auto_opt = record.get_option('AUTO')
            if auto_opt is not None and int(auto_opt) in [0, 1]:
                auto = bool(auto_opt)
            else:
                raise ValueError('Currently only AUTO=0 and AUTO=1 is supported')
        if record.has_option('PRINT'):
            keep_every_nth_iter = int(record.get_option('PRINT'))

        protected_names = [
            name.upper(),
            'EONLY',
            'INTERACTION',
            'INTER',
            'LAPLACE',
            'LAPLACIAN',
            'MAXEVAL',
            'MAXEVALS',
            'METHOD',
            'METH',
            'ISAMPLE',
            'NITER',
            'AUTO',
            'PRINT',
        ]

        tool_options = {
            option.key: option.value
            for option in record.all_options
            if option.key not in protected_names
        }
        if not tool_options:
            tool_options = None

        try:
            meth = EstimationStep.create(
                name,
                interaction=interaction,
                parameter_uncertainty_method=parameter_uncertainty_method,
                evaluation=evaluation,
                maximum_evaluations=maximum_evaluations,
                laplace=laplace,
                isample=isample,
                niter=niter,
                auto=auto,
                keep_every_nth_iter=keep_every_nth_iter,
                tool_options=tool_options,
                solver=solver,
                solver_rtol=tol,
                solver_atol=atol,
                eta_derivatives=etaderiv_names,
                epsilon_derivatives=epsilonderivs_names,
            )
        except ValueError:
            raise ModelSyntaxError(f'Non-recognized estimation method in: {str(record.root)}')
        steps.append(meth)

    steps = EstimationSteps.create(steps)

    return steps


def parse_solver(control_stream):
    subs_records = control_stream.get_records('SUBROUTINES')
    if not subs_records:
        return None, None, None
    record = subs_records[0]
    advan = record.advan
    # Currently only reading non-linear solvers
    # These can then be used if the model needs to use a non-linear solver
    if advan == 'ADVAN6':
        solver = 'DVERK'
    elif advan == 'ADVAN8':
        solver = 'DGEAR'
    elif advan == 'ADVAN9':
        solver = 'LSODI'
    elif advan == 'ADVAN13':
        solver = 'LSODA'
    elif advan == 'ADVAN14':
        solver = 'CVODES'
    elif advan == 'ADVAN15':
        solver = 'IDA'
    else:
        solver = None
    return solver, record.tol, record.atol


def parse_initial_individual_estimates(
    control_stream, name_map, basepath
) -> Optional[pd.DataFrame]:
    """Initial individual estimates

    These are taken from the $ETAS FILE.
    If no $ETAS is present None will be returned.

    Setter assumes that all IDs are present
    """
    etas = control_stream.get_records('ETAS')
    if etas:
        path = Path(etas[0].path)
        if not path.is_absolute():
            if basepath is None:
                raise ValueError("Cannot resolve path for $ETAS")
            path = path_absolute(basepath / path)
        try:
            phi_tables = NONMEMTableFile(path)
        except FileNotFoundError:
            return None
        phitab = phi_tables[0]
        assert isinstance(phitab, PhiTable)
        df = phitab.etas.rename(columns=name_map)
        return df
    else:
        return None


def parse_dataset_path(control_stream, basepath) -> Optional[Path]:
    record = next(iter(control_stream.get_records('DATA')), None)

    if record is None:
        return None

    path = Path(record.filename)
    if basepath is not None and not path.is_absolute():
        path = basepath.parent / path

    return path_absolute(path)


def _synonym(key, value):
    """Return a tuple reserved name and synonym"""
    _reserved_column_names = [
        'ID',
        'L1',
        'L2',
        'DV',
        'MDV',
        'RAW_',
        'MRG_',
        'RPT_',
        'TIME',
        'DATE',
        'DAT1',
        'DAT2',
        'DAT3',
        'EVID',
        'AMT',
        'RATE',
        'SS',
        'II',
        'ADDL',
        'CMT',
        'PCMT',
        'CALL',
        'CONT',
    ]
    if key in _reserved_column_names:
        return (key, value)
    elif value in _reserved_column_names:
        return (value, key)
    else:
        raise DatasetError(
            f'A column name "{key}" in $INPUT has a synonym to a non-reserved '
            f'column name "{value}"'
        )


def parse_column_info(control_stream):
    """List all column names in order.
    Use the synonym when synonym exists.
    return tuple of two lists, colnames, and drop together with a dictionary
    of replacements for reserved names (aka synonyms).
    Anonymous columns, i.e. DROP or SKIP alone, will be given unique names _DROP1, ...
    """
    input_records = control_stream.get_records("INPUT")
    colnames = []
    drop = []
    synonym_replacement = {}
    given_names = []
    next_anonymous = 1
    for record in input_records:
        for key, value in record.all_options:
            if value:
                if key == 'DROP' or key == 'SKIP':
                    colnames.append(value)
                    given_names.append(value)
                    drop.append(True)
                elif value == 'DROP' or value == 'SKIP':
                    colnames.append(key)
                    given_names.append(key)
                    drop.append(True)
                else:
                    (reserved_name, synonym) = _synonym(key, value)
                    synonym_replacement[reserved_name] = synonym
                    given_names.append(synonym)
                    colnames.append(synonym)
                    drop.append(False)
            else:
                if key == 'DROP' or key == 'SKIP':
                    name = f'_DROP{next_anonymous}'
                    next_anonymous += 1
                    colnames.append(name)
                    given_names.append(None)
                    drop.append(True)
                else:
                    colnames.append(key)
                    given_names.append(key)
                    drop.append(False)
    return colnames, drop, synonym_replacement, given_names


def parse_datainfo(control_stream, path) -> DataInfo:
    resolved_dataset_path = parse_dataset_path(control_stream, path)
    di_nonmem = create_nonmem_datainfo(control_stream, resolved_dataset_path)
    (colnames, drop, replacements, _) = parse_column_info(control_stream)

    if resolved_dataset_path is None:
        return di_nonmem

    dipath = resolved_dataset_path.with_suffix('.datainfo')
    if dipath.is_file():
        di_pharmpy = DataInfo.read_json(dipath)
        di_pharmpy = di_pharmpy.replace(path=resolved_dataset_path)
        di_pharmpy = validate_datainfo(di_pharmpy, di_nonmem)
        return di_pharmpy
    return di_nonmem


def validate_datainfo(di_pharmpy, di_nonmem):
    different_drop, cols_new = [], []
    for col_pharmpy, col_nonmem in zip(di_pharmpy, di_nonmem):
        if col_pharmpy.drop != col_nonmem.drop:
            colinfo_new = col_pharmpy.replace(drop=col_nonmem.drop)
            cols_new.append(colinfo_new)
            if col_nonmem.datatype != 'nmtran-date':
                different_drop.append(col_pharmpy.name)
        else:
            cols_new.append(col_pharmpy)

    if cols_new:
        di_new = di_pharmpy.replace(columns=tuple(cols_new))
        if different_drop:
            warnings.warn(
                "NONMEM .mod and dataset .datainfo disagree on "
                f"DROP for columns {', '.join(different_drop)}."
            )
        return di_new
    return di_pharmpy


def create_nonmem_datainfo(control_stream, resolved_dataset_path):
    (colnames, drop, replacements, _) = parse_column_info(control_stream)

    column_info = []
    have_pk = control_stream.get_pk_record()
    for colname, coldrop in zip(colnames, drop):
        if coldrop and colname not in ['DATE', 'DAT1', 'DAT2', 'DAT3']:
            info = ColumnInfo.create(colname, drop=coldrop, datatype='str')
        elif colname == 'ID' or colname == 'L1':
            info = ColumnInfo.create(
                colname, drop=coldrop, datatype='int32', type='id', scale='nominal'
            )
        elif colname == 'DV' or colname == replacements.get('DV', None):
            info = ColumnInfo.create(colname, drop=coldrop, type='dv')
        elif colname == 'TIME' or colname == replacements.get('TIME', None):
            if not set(colnames).isdisjoint({'DATE', 'DAT1', 'DAT2', 'DAT3'}):
                datatype = 'nmtran-time'
            else:
                datatype = 'float64'
            info = ColumnInfo.create(
                colname, drop=coldrop, type='idv', scale='ratio', datatype=datatype
            )
        elif colname in ['DATE', 'DAT1', 'DAT2', 'DAT3']:
            # Always DROP in mod-file, but actually always used
            info = ColumnInfo.create(colname, drop=False, scale='interval', datatype='nmtran-date')
        elif colname == 'EVID' and have_pk:
            info = ColumnInfo.create(colname, drop=coldrop, type='event', scale='nominal')
        elif colname == 'MDV' and have_pk:
            if 'EVID' in colnames:
                tp = 'mdv'
            else:
                tp = 'event'
            info = ColumnInfo.create(
                colname, drop=coldrop, type=tp, scale='nominal', datatype='int32'
            )
        elif colname == 'II' and have_pk:
            info = ColumnInfo.create(colname, drop=coldrop, type='ii', scale='ratio')
        elif colname == 'SS' and have_pk:
            info = ColumnInfo.create(colname, drop=coldrop, type='ss', scale='nominal')
        elif colname == 'ADDL' and have_pk:
            info = ColumnInfo.create(colname, drop=coldrop, type='additional', scale='ordinal')
        elif (colname == 'AMT' or colname == replacements.get('AMT', None)) and have_pk:
            info = ColumnInfo.create(colname, drop=coldrop, type='dose', scale='ratio')
        elif colname == 'CMT' and have_pk:
            info = ColumnInfo.create(colname, drop=coldrop, type='compartment', scale='nominal')
        elif colname == 'RATE' and have_pk:
            info = ColumnInfo.create(colname, drop=coldrop, type='rate')
        elif colname == 'BLQ':
            info = ColumnInfo.create(colname, drop=coldrop, type='blq', scale='nominal')
        elif colname == 'LLOQ':
            info = ColumnInfo.create(colname, drop=coldrop, type='lloq')
        else:
            info = ColumnInfo.create(colname, drop=coldrop)
        column_info.append(info)

    di = DataInfo.create(column_info, path=resolved_dataset_path)
    return di


def replace_synonym_in_filters(filters, replacements):
    result = []
    for f in filters:
        col = f.leaf('COLUMN').value
        if col in replacements:
            s = ''
            for child in f.children:
                if child.rule == 'COLUMN':
                    value = replacements[col]
                else:
                    value = str(child)
                s += value
        else:
            s = str(f)
        result.append(s)
    return result


def parse_dataset(
    di: DataInfo,
    control_stream: NMTranControlStream,
    raw: bool = False,
    parse_columns: Tuple[str, ...] = (),
):
    data_records = control_stream.get_records('DATA')
    if not data_records:
        return None
    ignore_character = data_records[0].ignore_character
    null_value = data_records[0].null_value
    (colnames, drop, replacements, _) = parse_column_info(control_stream)

    if raw:
        ignore = None
        accept = None
    else:
        # FIXME: All direct handling of control stream spanning
        # over one or more records should move
        ignore = data_records[0].ignore
        accept = data_records[0].accept
        # FIXME: This should really only be done if setting the dataset
        if ignore:
            ignore = replace_synonym_in_filters(ignore, replacements)
        else:
            accept = replace_synonym_in_filters(accept, replacements)

    df = read_nonmem_dataset(
        di.path,
        raw,
        ignore_character,
        colnames,
        drop,
        null_value=null_value,
        parse_columns=parse_columns,
        ignore=ignore,
        accept=accept,
        dtype=None if raw else di.get_dtype_dict(),
    )
    # Let TIME be the idv in both $PK and $PRED models
    # Remove individuals without observations
    col_names = list(df.columns)
    have_pk = control_stream.get_pk_record()
    if have_pk:
        df = filter_observations(df, col_names)
    return df


def filter_observations(df, col_names):
    if 'EVID' in col_names:
        df_obs = df.astype({'EVID': 'float'}).query('EVID == 0')
    elif 'MDV' in col_names:
        df_obs = df.astype({'MDV': 'float'}).query('MDV == 0')
    elif 'AMT' in col_names:
        df_obs = df.astype({'AMT': 'float'}).query('AMT == 0')
    else:
        raise DatasetError('Could not identify observation rows in dataset')
    have_obs = set(df_obs['ID'].unique())
    all_ids = set(df['ID'].unique())
    ids_to_remove = all_ids - have_obs
    return df[~df['ID'].isin(ids_to_remove)]


def parse_table_columns(control_stream, netas):
    # Handle synonyms and appended columns

    reserved = {'PRED', 'IPRED', 'CIPREDI'}
    synonyms = dict()
    all_columns = []

    code_recs = (
        control_stream.get_records('PK')
        + control_stream.get_records('PRED')
        + control_stream.get_records('ERROR')
    )
    symbs = set()
    for rec in code_recs:
        for s in rec.statements:
            symbs.add(s.symbol.name)

    (colnames, _, _, _) = parse_column_info(control_stream)
    symbs |= set(colnames)

    for table_record in control_stream.get_records('TABLE'):
        noappend = False
        columns = []
        for opt, key, value in table_record.parse_options(nonoptions=symbs, netas=netas):
            if key == 'NOAPPEND':
                noappend = True
            elif opt.need_value is None:
                if value is None:
                    if key in synonyms:
                        columns.append(synonyms[key])
                    else:
                        columns.append(key)
                else:
                    if key in reserved:
                        columns.append(key)
                        synonyms[value] = key
                    elif value in reserved:
                        columns.append(value)
                        synonyms[key] = value
                    else:
                        # FIXME: Fallback since we don't know all reserved names
                        columns.append(key)

        if not noappend:
            toremove = ['PRED', 'RES', 'WRES']
            toappend = ['DV'] + toremove
            # Remove appended columns explicitly in $TABLE except DV
            columns = [col for col in columns if col not in toremove]
            columns.extend(toappend)

        all_columns.append(columns)

    return all_columns



import numpy as np
import pytest
import sympy
from sympy import Symbol as symbol

from pharmpy.model import Parameter, Parameters


@pytest.mark.parametrize(
    'name,init,lower,upper,fix',
    [
        ('THETA(1)', 23, None, None, None),
        ('X', 12, None, None, None),
        ('_NAME', 0, None, None, False),
        ('OMEGA(2,1)', 0.1, 0, None, None),
        ('TCVL', 0.23, -2, 2, None),
    ],
)
def test_initialization(name, init, lower, upper, fix):
    if lower is None:
        lower = -sympy.oo
    if upper is None:
        upper = sympy.oo
    if fix is None:
        fix = False
    param = Parameter(name, init, lower, upper, fix)
    assert param.name == name
    assert param.symbol == symbol(name)
    assert param.init == init
    assert param.lower == lower
    assert param.upper == upper
    assert param.fix == bool(fix)


@pytest.mark.parametrize(
    'name,init,lower,upper,fix',
    [
        ('OMEGA(2,1)', 0.1, 2, None, None),
        ('X', 1, 0, -1, None),
        ('X', np.nan, 0, 2, None),
        (23, 0, None, None, None),
    ],
)
def test_illegal_initialization(name, init, lower, upper, fix):
    with pytest.raises(ValueError):
        Parameter.create(name, init, lower, upper, fix)


def test_repr():
    param = Parameter('X', 2, lower=0, upper=23)
    assert repr(param) == 'Parameter("X", 2, lower=0, upper=23, fix=False)'
    param = Parameter('X', 2)
    assert repr(param) == 'Parameter("X", 2, lower=-∞, upper=∞, fix=False)'


def test_fix():
    param = Parameter('X', 2, lower=0, upper=23)
    assert param == Parameter('X', 2, lower=0, upper=23, fix=False)
    param = param.replace(fix=True)
    assert param == Parameter('X', 2, lower=0, upper=23, fix=True)


def test_add():
    p1 = Parameter.create('x', 2)
    p2 = Parameter.create('y', 3)
    p3 = Parameter.create('z', 4)
    pset = Parameters((p1, p2))
    pset2 = pset + p3
    assert len(pset2) == 3
    assert pset2['z'].init == 4
    pset3 = Parameters((p1,))
    pset4 = pset3 + Parameters((p2, p3))
    assert len(pset4) == 3
    assert pset4['z'].init == 4
    pset5 = Parameters((p1,))
    pset6 = pset5 + (p2, p3)
    assert len(pset6) == 3
    assert pset6['z'].init == 4

    with pytest.raises(ValueError):
        pset + 23


def test_pset_radd():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset1 = Parameters((p1, p2, p3))
    p4 = Parameter.create('W', 1)
    cat = p4 + pset1
    assert len(cat) == 4
    assert cat['W'].init == 1
    assert cat['Y'].init == 9

    cat = [p4] + pset1
    assert len(cat) == 4
    assert cat['W'].init == 1
    assert cat['Y'].init == 9

    with pytest.raises(ValueError):
        23 + pset1


def test_pset_init():
    p = Parameter.create('Y', 9)
    pset = Parameters((p,))
    pset2 = Parameters.create(pset)
    assert len(pset2) == 1
    assert pset2['Y'].init == 9

    p2 = Parameter.create('Y', 12)
    with pytest.raises(ValueError):
        Parameters.create((p, p2))

    with pytest.raises(ValueError):
        Parameters.create([23])

    pset3 = Parameters.create()
    assert len(pset3) == 0


def test_pset_getitem():
    p = Parameter.create('Y', 9)
    pset = Parameters((p,))
    assert len(pset) == 1
    assert pset['Y'] is p

    p2 = Parameter.create('Z', 5)
    pset = Parameters((p, p2))

    assert len(pset) == 2

    # Check that the parameter set keeps the insertion order upon iteration
    for i, param in enumerate(pset):
        if i == 0:
            assert param is p
        else:
            assert param is p2

    assert pset[symbol('Z')] == p2

    p3 = Parameter.create('K', 19)
    with pytest.raises(KeyError):
        pset[p3]

    assert len(pset[[p]]) == 1

    with pytest.raises(KeyError):
        pset['noparamofmine']


def test_pset_nonfixed():
    p1 = Parameter.create('Y', 9, fix=False)
    p2 = Parameter.create('X', 3, fix=True)
    p3 = Parameter.create('Z', 1, fix=False)
    pset = Parameters((p1, p2, p3))
    assert len(pset.nonfixed) == 2
    assert pset.nonfixed['Y'] == Parameter.create('Y', 9)


def test_pset_names():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset = Parameters((p1, p2, p3))
    assert pset.names == ['Y', 'X', 'Z']
    assert pset.symbols == [symbol('Y'), symbol('X'), symbol('Z')]


def test_pset_lower_upper():
    p1 = Parameter('X', 0, lower=-1, upper=1)
    p2 = Parameter.create('Y', 1, lower=0)
    pset = Parameters((p1, p2))
    assert pset.lower == {'X': -1, 'Y': 0}
    assert pset.upper == {'X': 1, 'Y': sympy.oo}


def test_pset_nonfixed_inits():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset = Parameters((p1, p2, p3))
    assert pset.nonfixed.inits == {'Y': 9, 'X': 3, 'Z': 1}


def test_pset_fix():
    p1 = Parameter.create('Y', 9, fix=False)
    p2 = Parameter.create('X', 3, fix=True)
    p3 = Parameter.create('Z', 1, fix=False)
    pset = Parameters((p1, p2, p3))
    assert pset.fix == {'Y': False, 'X': True, 'Z': False}


def test_pset_repr():
    p1 = Parameter.create('Y', 9, fix=False)
    pset = Parameters((p1,))
    assert type(repr(pset)) == str
    assert type(pset._repr_html_()) == str
    pset = Parameters()
    assert type(repr(pset)) == str
    assert type(pset._repr_html_()) == str


def test_pset_eq():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset1 = Parameters((p1, p2, p3))
    pset2 = Parameters((p1, p2))
    assert pset1 != pset2
    pset3 = Parameters((p1, p3, p2))
    assert pset1 != pset3
    assert pset1 == pset1
    assert pset1 != 23


def test_pset_replace():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset1 = Parameters((p1, p2, p3))
    pset2 = pset1.replace(parameters=(p1, p2))
    assert len(pset2) == 2


def test_hash():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('Y', 9, upper=23)
    assert hash(p1) != hash(p2)
    p3 = Parameter.create('X', 9)
    p4 = Parameter.create('Z', 9)
    pset1 = Parameters((p1, p3))
    pset2 = Parameters((p1, p4))
    assert hash(pset1) != hash(pset2)


def test_dict():
    p1 = Parameter.create('Y', 9)
    d = p1.to_dict()
    assert d == {
        'name': 'Y',
        'init': 9.0,
        'lower': -float('inf'),
        'upper': float('inf'),
        'fix': False,
    }
    p2 = Parameter.from_dict(d)
    assert p1 == p2
    p3 = Parameter.create('Z', 0, lower=0, upper=22)
    pset = Parameters((p1, p3))
    d2 = pset.to_dict()
    assert d2 == {
        'parameters': (
            {'name': 'Y', 'init': 9.0, 'lower': -float('inf'), 'upper': float('inf'), 'fix': False},
            {'name': 'Z', 'init': 0.0, 'lower': 0.0, 'upper': 22.0, 'fix': False},
        )
    }
    pset2 = Parameters.from_dict(d2)
    assert pset == pset2


def test_contains():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset1 = Parameters((p1, p2, p3))
    assert 'Y' in pset1
    assert 'Q' not in pset1


def test_set_initial_estimates():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset1 = Parameters((p1, p2, p3))
    pset2 = pset1.set_initial_estimates({'Y': 23, 'Z': 19})
    assert pset2['Y'].init == 23
    assert pset2['X'].init == 3
    assert pset2['Z'].init == 19
    assert pset1['Y'].init == 9
    assert pset1['X'].init == 3
    assert pset1['Z'].init == 1


def test_set_fix():
    p1 = Parameter.create('Y', 9, fix=True)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset1 = Parameters((p1, p2, p3))
    pset2 = pset1.set_fix({'Y': False, 'X': True})
    assert not pset2['Y'].fix
    assert pset2['X'].fix
    assert not pset2['Z'].fix
    assert pset1['Y'].fix
    assert not pset1['X'].fix
    assert not pset1['Z'].fix


def test_fixed_nonfixed():
    p1 = Parameter.create('Y', 9, fix=True)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset1 = Parameters((p1, p2, p3))
    pset_fixed = Parameters((p1,))
    pset_nonfixed = Parameters((p2, p3))
    assert pset1.fixed == pset_fixed
    assert pset1.nonfixed == pset_nonfixed


def test_slice():
    p1 = Parameter.create('Y', 9)
    p2 = Parameter.create('X', 3)
    p3 = Parameter.create('Z', 1)
    pset1 = Parameters((p1, p2, p3))
    sl = pset1[1:]
    assert len(sl) == 2
    assert sl['X'].init == 3
    assert 'Y' not in sl


def test_replace():
    p = Parameter.create('x', 3, fix=False)
    p1 = p.replace(upper=4)
    assert not p1.fix
    assert not p.fix
    assert p1.name == 'x'
    assert p1.upper == 4

    p2 = p.replace(lower=2)
    assert p2.lower == 2

    p3 = p.replace(init=23)
    assert p3.init == 23



import json
import re

from pharmpy.model import Model


def detect_model(src):
    is_generic = re.search(r'"__magic__"\s*:\s*"Pharmpy Model"', src, re.MULTILINE)
    return is_generic


def convert_model(model):
    new = Model(
        dataset=model.dataset,
        datainfo=model.datainfo,
        name=model.name,
        parameters=model.parameters,
        statements=model.statements,
        random_variables=model.random_variables,
        estimation_steps=model.estimation_steps,
        dependent_variables=model.dependent_variables,
        observation_transformation=model.observation_transformation,
        description=model.description,
        parent_model=model.name,
        filename_extension=model.filename_extension,
        initial_individual_estimates=model.initial_individual_estimates,
    )
    return new


def parse_model(code, path=None):
    d = json.loads(code)
    model = Model.from_dict(d)
    return model



def optimize_task_graph_for_dask_distributed(client, graph):
    from dask.distributed import Future

    optimized = {key: _scatter_computation(Future, client, value) for key, value in graph.items()}
    from dask.optimization import fuse

    return fuse(optimized)[0]


def _scatter_computation(Future, client, computation):
    # NOTE: According to dask's graph spec (https://docs.dask.org/en/stable/spec.html):
    # A computation may be one of the following:
    #  - Any key present in the Dask graph like 'x'
    #  - Any other value like 1, to be interpreted literally
    #  - A task like (inc, 'x') (see below)
    #  - A list of computations, like [1, 'x', (inc, 'x')]

    if isinstance(computation, tuple):
        return (
            computation[0],
            *map(lambda c: _scatter_computation(Future, client, c), computation[1:]),
        )

    if isinstance(computation, list):
        return list(map(lambda c: _scatter_computation(Future, client, c), computation))

    return _scatter_value(Future, client, computation)


def _scatter_value(Future, client, value):
    # TODO: We could automatically compute whether object size is above
    # threshold with a slight twist on https://stackoverflow.com/a/30316760
    if isinstance(value, (int, str, float, bool, range, Future)) or callable(value):
        return value
    else:
        return client.scatter(value)



from __future__ import annotations

import importlib.util
import warnings
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from pharmpy.model import Model, ModelfitResultsError
from pharmpy.workflows import ModelfitResults

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import pandas as pd

from .ml import predict_influential_individuals, predict_outliers


def summarize_individuals(models: List[Model], models_res: List[ModelfitResults]) -> pd.DataFrame:
    """Creates a summary dataframe keyed by model-individual pairs for an input
    list of models.

    Content of the various columns:

    +-------------------------+----------------------------------------------------------------------+
    | Column                  | Description                                                          |
    +=========================+======================================================================+
    | ``outlier_count``       | Number of observations with CWRES > 5                                |
    +-------------------------+----------------------------------------------------------------------+
    | ``ofv``                 | Individual OFV                                                       |
    +-------------------------+----------------------------------------------------------------------+
    | ``dofv_vs_parent``      | Difference in individual OFV between this model and its parent model |
    +-------------------------+----------------------------------------------------------------------+
    | ``predicted_dofv``      | Predicted dOFV if this individual was excluded                       |
    +-------------------------+----------------------------------------------------------------------+
    | ``predicted_residual``  | Predicted residual                                                   |
    +-------------------------+----------------------------------------------------------------------+

    Parameters
    ----------
    models : List[Model]
        Input models
    models_res : List[ModelfitResults]
        Input results

    Return
    ------
    pd.DataFrame | None
        The summary as a dataframe

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> from pharmpy.tools import fit, summarize_individuals
    >>> fit_results = fit(model)  # doctest: +SKIP
    <Pharmpy model object pheno>
    >>> from pharmpy.tools import run_tool # doctest: +SKIP
    >>> results = run_tool(
    ...     'modelsearch',
    ...     model=model,
    ...     mfl='ABSORPTION(ZO);PERIPHERALS([1, 2])',
    ...     algorithm='reduced_stepwise'
    ... ) # doctest: +SKIP
    >>> summarize_individuals([results.start_model, *results.models]) # doctest: +SKIP

    """  # noqa: E501
    if len(models) != len(models_res):
        raise ValueError('Different length of `models` and `models_res`')

    resDict = {model.name: res for model, res in zip(models, models_res)}

    spec = importlib.util.find_spec('tflite_runtime')
    if spec is None:
        warnings.warn("tflite is not installed, using NaN for predictions")

    df = pd.concat(
        map(
            lambda model: groupedByIDAddColumnsOneModel(resDict, model, resDict[model.name]),
            models,
        ),
        keys=[model.name for model in models],
        names=['model'],
        axis=0,
        verify_integrity=True,
    )

    assert df is not None
    return df


def parent_model_name(model: Model) -> str:
    name = model.parent_model
    assert isinstance(name, str)
    return name


def model_name(model: Model) -> str:
    return model.name


def outlier_count_func(df: pd.DataFrame) -> float:
    # NOTE: This returns a float because we will potentially concat this
    # with NaNs
    return float((abs(df) > 5).sum())


def outlier_count(res: ModelfitResults) -> Union[pd.Series, float]:
    if res is None:
        return np.nan
    residuals = res.residuals
    if residuals is None:
        return np.nan
    else:
        groupedByID = residuals.groupby('ID')
        return groupedByID['CWRES'].agg(outlier_count_func)


def _predicted(predict, model: Model, res: ModelfitResults, column: str) -> Union[pd.Series, float]:
    try:
        predicted = predict(model, res)
    except ModelfitResultsError:
        return np.nan
    except ImportError:
        return np.nan
    if predicted is None:
        return np.nan
    return predicted[column]


def predicted_residual(model: Model, res: ModelfitResults) -> Union[pd.Series, float]:
    return _predicted(predict_outliers, model, res, 'residual')


def predicted_dofv(model: Model, res: ModelfitResults) -> Union[pd.Series, float]:
    return _predicted(predict_influential_individuals, model, res, 'dofv')


def ofv(res: ModelfitResults) -> Union[pd.Series, float]:
    return np.nan if res is None or res.individual_ofv is None else res.individual_ofv


def dofv(
    parent_model_res: Optional[ModelfitResults], candidate_model_res: Optional[ModelfitResults]
) -> Union[pd.Series, float]:
    return np.nan if parent_model_res is None else ofv(parent_model_res) - ofv(candidate_model_res)


def groupedByIDAddColumnsOneModel(
    resDict: Dict[str, ModelfitResults], model: Model, model_res: ModelfitResults
) -> pd.DataFrame:
    id_column_name = model.datainfo.id_column.name
    index = pd.Index(data=model.dataset[id_column_name].unique(), name=id_column_name)
    parent_model_name = model.parent_model
    parent_model_res = None if parent_model_name is None else resDict.get(parent_model_name)
    df = pd.DataFrame(
        {
            'parent_model': parent_model_name,
            'outlier_count': outlier_count(model_res),
            'ofv': ofv(model_res),
            'dofv_vs_parent': dofv(parent_model_res, model_res),
            'predicted_dofv': predicted_dofv(model, model_res),
            'predicted_residual': predicted_residual(model, model_res),
        },
        index=index,
    )
    return df


def summarize_individuals_count_table(
    models: Optional[List[Model]] = None,
    models_res: Optional[List[ModelfitResults]] = None,
    df: pd.DataFrame = None,
):
    r"""Create a count table for individual data

    Content of the various columns:

    +-------------------------+------------------------------------------------------------------------------------------------+
    | Column                  | Description                                                                                    |
    +=========================+================================================================================================+
    | ``inf_selection``       | Number of subjects influential on model selection.                                             |
    |                         | :math:`\mathrm{OFV}_{parent} - \mathrm{OFV} > 3.84 \veebar`                                    |
    |                         | :math:`\mathrm{OFV}_{parent} - \mathrm{iOFV}_{parent} - (\mathrm{OFV} - \mathrm{iOFV}) > 3.84` |
    +-------------------------+------------------------------------------------------------------------------------------------+
    | ``inf_params``          | Number of subjects influential on parameters. predicted_dofv > 3.84                            |
    +-------------------------+------------------------------------------------------------------------------------------------+
    | ``out_obs``             | Number of subjects having at least one outlying observation (CWRES > 5)                        |
    +-------------------------+------------------------------------------------------------------------------------------------+
    | ``out_ind``             | Number of outlying subjects. predicted_residual > 3.0                                          |
    +-------------------------+------------------------------------------------------------------------------------------------+
    | ``inf_outlier``         | Number of subjects both influential by any criteria and outlier by any criteria                |
    +-------------------------+------------------------------------------------------------------------------------------------+

    Parameters
    ----------
    models : list of models
        List of models to summarize.
    models_res : List[ModelfitResults]
        Input results
    df : pd.DataFrame
        Output from a previous call to summarize_individuals.

    Returns
    -------
    pd.DataFrame
        Table with one row per model.

    See also
    --------
    summarize_individuals : Get raw individual data

    """  # noqa: E501
    if models and models_res:
        if len(models) != len(models_res):
            raise ValueError('Different length of `models` and `models_res`')
        df = summarize_individuals(models, models_res)
    if df is None:
        return None

    is_out_obs = df['outlier_count'] > 0.0
    is_out_ind = df['predicted_residual'] > 3.0
    is_inf_params = df['predicted_dofv'] > 3.84

    out_obs = is_out_obs.groupby(level='model', sort=False).sum().astype('int32')
    out_ind = is_out_ind.groupby(level='model', sort=False).sum().astype('int32')
    inf_params = is_inf_params.groupby(level='model', sort=False).sum().astype('int32')

    ninds = len(df.index.unique(level='ID'))
    parents = df['parent_model'].iloc[::ninds]
    parent_ofvs = df.loc[parents]['ofv'].reset_index(drop=True)
    parent_ofvs.index = df.index

    for name in df.index.unique(level='model'):
        if name == df.loc[name]['parent_model'].iloc[0]:
            start_name = name
            break
    else:
        # FIXME: Handle missing start model
        raise ValueError('Missing start model')

    ofv_sums = df['ofv'].groupby('model').sum()
    parent_sums = parent_ofvs.groupby('model').sum()
    full_ofv_diff = parent_sums - ofv_sums  # / len(df.index.unique(level='ID'))
    full_ofv_diff.loc[start_name] = 0

    removed_diff = (parent_sums - parent_ofvs) - (ofv_sums - df['ofv'])
    is_inf_selection = (full_ofv_diff > 3.84) ^ (removed_diff > 3.84)
    inf_selection = is_inf_selection.groupby(level='model', sort=False).sum().astype('int32')

    is_inf_outlier = (is_out_obs | is_out_ind) & (is_inf_params | is_inf_selection)
    inf_outlier = is_inf_outlier.groupby(level='model', sort=False).sum().astype('int32')
    parents.index = inf_selection.index
    res = pd.DataFrame(
        {
            'parent_model': parents,
            'inf_selection': inf_selection,
            'inf_params': inf_params,
            'out_obs': out_obs,
            'out_ind': out_ind,
            'inf_outlier': inf_outlier,
        }
    )
    return res



from __future__ import annotations

import warnings
from dataclasses import dataclass, replace
from typing import List, Literal, Optional, Union

import pharmpy.tools.iivsearch.algorithms as algorithms
from pharmpy.deps import pandas as pd
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.model import Model
from pharmpy.modeling import (
    add_pd_iiv,
    add_pk_iiv,
    calculate_bic,
    create_joint_distribution,
    has_random_effect,
)
from pharmpy.tools import summarize_modelfit_results
from pharmpy.tools.common import RANK_TYPES, ToolResults, create_results, update_initial_estimates
from pharmpy.tools.iivsearch.algorithms import _get_fixed_etas
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder, call_workflow
from pharmpy.workflows.results import ModelfitResults

IIV_STRATEGIES = frozenset(
    ('no_add', 'add_diagonal', 'fullblock', 'pd_add_diagonal', 'pd_fullblock')
)
IIV_ALGORITHMS = frozenset(('brute_force', 'brute_force_no_of_etas', 'brute_force_block_structure'))


def create_workflow(
    algorithm: Literal[tuple(IIV_ALGORITHMS)],
    iiv_strategy: Literal[tuple(IIV_STRATEGIES)] = 'no_add',
    rank_type: Literal[tuple(RANK_TYPES)] = 'mbic',
    cutoff: Optional[Union[float, int]] = None,
    results: Optional[ModelfitResults] = None,
    model: Optional[Model] = None,
    keep: Optional[List[str]] = None,
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs>=0.1)",
):
    """Run IIVsearch tool. For more details, see :ref:`iivsearch`.

    Parameters
    ----------
    algorithm : {'brute_force', 'brute_force_no_of_etas', 'brute_force_block_structure'}
        Which algorithm to run.
    iiv_strategy : {'no_add', 'add_diagonal', 'fullblock', 'pd_add_diagonal', 'pd_fullblock'}
        If/how IIV should be added to start model. Default is 'no_add'.
    rank_type : {'ofv', 'lrt', 'aic', 'bic', 'mbic'}
        Which ranking type should be used. Default is mBIC.
    cutoff : float
        Cutoff for which value of the ranking function that is considered significant. Default
        is None (all models will be ranked)
    results : ModelfitResults
        Results for model
    model : Model
        Pharmpy model
    keep :  List[str]
        List of IIVs to keep
    strictness : str or None
        Strictness criteria

    Returns
    -------
    IIVSearchResults
        IIVsearch tool result object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import run_iivsearch, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> run_iivsearch('brute_force', results=results, model=model)   # doctest: +SKIP
    """

    wb = WorkflowBuilder(name='iivsearch')
    start_task = Task(
        'start_iiv',
        start,
        model,
        results,
        algorithm,
        iiv_strategy,
        rank_type,
        cutoff,
        keep,
        strictness,
    )
    wb.add_task(start_task)
    task_results = Task('results', _results)
    wb.add_task(task_results, predecessors=[start_task])
    return Workflow(wb)


def create_step_workflow(
    input_model_entry, base_model_entry, wf_algorithm, iiv_strategy, rank_type, cutoff, strictness
):
    wb = WorkflowBuilder()
    start_task = Task(f'start_{wf_algorithm.name}', _start_algorithm, base_model_entry)
    wb.add_task(start_task)

    if iiv_strategy != 'no_add':
        wf_fit = create_fit_workflow(n=1)
        wb.insert_workflow(wf_fit)
        base_model_task = wf_fit.output_tasks[0]
    else:
        base_model_task = start_task

    wb.insert_workflow(wf_algorithm)

    task_result = Task(
        'results',
        post_process,
        rank_type,
        cutoff,
        strictness,
        input_model_entry,
        base_model_entry.model.name,
    )

    post_process_tasks = [base_model_task] + wb.output_tasks
    wb.add_task(task_result, predecessors=post_process_tasks)

    return Workflow(wb)


def start(
    context, input_model, input_res, algorithm, iiv_strategy, rank_type, cutoff, keep, strictness
):
    input_model_entry = ModelEntry.create(input_model, modelfit_results=input_res)

    if iiv_strategy != 'no_add':
        base_model = update_initial_estimates(input_model, modelfit_results=input_res)
        base_model = _add_iiv(iiv_strategy, base_model, modelfit_results=input_res)
        base_model = base_model.replace(
            name='base_model', description=algorithms.create_description(base_model)
        )
        # FIXME: Set parent model once create_results fully supports model entries
        base_model_entry = ModelEntry.create(base_model, modelfit_results=None)
    else:
        base_model_entry = ModelEntry.create(input_model, modelfit_results=input_res)

    if algorithm == 'brute_force':
        list_of_algorithms = ['brute_force_no_of_etas', 'brute_force_block_structure']
    else:
        list_of_algorithms = [algorithm]
    sum_tools, sum_models, sum_inds, sum_inds_count, sum_errs = [], [], [], [], []

    no_of_models = 0
    last_res = None
    final_model_entry = None

    sum_models = [summarize_modelfit_results(input_res)]

    for algorithm_cur in list_of_algorithms:
        algorithm_func = getattr(algorithms, algorithm_cur)
        if algorithm_cur == "brute_force_no_of_etas":
            # NOTE: This does not need to be a model entry since it is only used as a start point for the
            # candidate models, when the workflow is run the input to this sub-workflow will be a model entry
            wf_algorithm = algorithm_func(
                base_model_entry.model, index_offset=no_of_models, keep=keep
            )
        else:
            wf_algorithm = algorithm_func(base_model_entry.model, index_offset=no_of_models)

        wf = create_step_workflow(
            input_model_entry,
            base_model_entry,
            wf_algorithm,
            iiv_strategy,
            rank_type,
            cutoff,
            strictness,
        )
        res = call_workflow(wf, f'results_{algorithm}', context)

        if base_model_entry.model.name in sum_models[-1].index.values:
            summary_models = res.summary_models.drop(base_model_entry.model.name, axis=0)
        else:
            summary_models = res.summary_models

        sum_tools.append(res.summary_tool)
        sum_models.append(summary_models)
        sum_inds.append(res.summary_individuals)
        sum_inds_count.append(res.summary_individuals_count)
        sum_errs.append(res.summary_errors)

        if res.final_model.name != input_model_entry.model.name:
            final_model_entry = context.model_database.retrieve_model_entry(res.final_model.name)
        else:
            final_model_entry = input_model_entry
        base_model_entry = final_model_entry
        iiv_strategy = 'no_add'
        last_res = res
        no_of_models = len(res.summary_tool) - 1

        assert base_model_entry is not None
        if (
            len(
                set(base_model_entry.model.random_variables.iiv.names).difference(
                    _get_fixed_etas(base_model_entry.model)
                )
            )
            <= 1
        ):
            break

    assert last_res is not None
    assert final_model_entry is not None

    input_model, input_res = input_model_entry.model, input_model_entry.modelfit_results
    final_model, final_res = final_model_entry.model, final_model_entry.modelfit_results

    # NOTE: Compute final final model
    final_final_model = last_res.final_model
    if input_res and final_res:
        bic_input = calculate_bic(input_model, input_res.ofv, type='iiv')
        bic_final = calculate_bic(final_model, final_res.ofv, type='iiv')
        if bic_final > bic_input:
            warnings.warn(
                f'Worse {rank_type} in final model {final_model.name} '
                f'({bic_final}) than {input_model.name} ({bic_input}), selecting '
                f'input model'
            )
            final_final_model = input_model

    keys = list(range(1, len(list_of_algorithms) + 1))

    return IIVSearchResults(
        summary_tool=_concat_summaries(sum_tools, keys),
        summary_models=_concat_summaries(sum_models, [0] + keys),  # To include input model
        summary_individuals=_concat_summaries(sum_inds, keys),
        summary_individuals_count=_concat_summaries(sum_inds_count, keys),
        summary_errors=_concat_summaries(sum_errs, keys),
        final_model=final_final_model,
        tool_database=last_res.tool_database,
    )


def _concat_summaries(summaries, keys):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="The behavior of DataFrame concatenation with empty or all-NA",
            category=FutureWarning,
        )
        return pd.concat(summaries, keys=keys, names=['step'])


def _results(res):
    return res


def _start_algorithm(model_entry):
    return model_entry


def _add_iiv(iiv_strategy, model, modelfit_results):
    assert iiv_strategy in ['add_diagonal', 'fullblock', 'pd_add_diagonal', 'pd_fullblock']
    if iiv_strategy in ['add_diagonal', 'fullblock']:
        model = add_pk_iiv(model)
        if iiv_strategy == 'fullblock':
            model = create_joint_distribution(
                model, individual_estimates=modelfit_results.individual_estimates
            )
    elif iiv_strategy in ['pd_add_diagonal', 'pd_fullblock']:
        model = add_pd_iiv(model)
        if iiv_strategy == 'pd_fullblock':
            model = create_joint_distribution(
                model, individual_estimates=modelfit_results.individual_estimates
            )
    return model


def post_process(rank_type, cutoff, strictness, input_model_entry, base_model_name, *model_entries):
    res_model_entries = []
    base_model_entry = None
    for model_entry in model_entries:
        if model_entry.model.name == base_model_name:
            base_model_entry = model_entry
        else:
            res_model_entries.append(model_entry)

    assert len(res_model_entries) > 0

    if not base_model_entry:
        raise ValueError('Error in workflow: No base model')

    # In order to have the IIV structure of the input model in the description column
    # in the result summaries
    if input_model_entry.model.name == base_model_entry.model.name:
        base_model = base_model_entry.model
        base_model_description = algorithms.create_description(base_model)
        base_model = base_model.replace(description=base_model_description)
        base_model_entry = ModelEntry.create(
            base_model, modelfit_results=base_model_entry.modelfit_results
        )

    res = create_results(
        IIVSearchResults,
        input_model_entry,
        base_model_entry,
        res_model_entries,
        rank_type,
        cutoff,
        bic_type='iiv',
        strictness=strictness,
    )

    summary_tool = res.summary_tool
    assert summary_tool is not None
    summary_models = summarize_modelfit_results(
        [model_entry.modelfit_results for model_entry in model_entries]
    )

    return replace(res, summary_models=summary_models)


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(
    algorithm,
    iiv_strategy,
    rank_type,
    model,
    keep,
    strictness,
):
    if keep:
        for parameter in keep:
            try:
                has_random_effect(model, parameter, "iiv")
            except KeyError:
                raise ValueError(f"Parameter {parameter} has no iiv.")

    if strictness is not None and "rse" in strictness.lower():
        if model.estimation_steps[-1].parameter_uncertainty_method is None:
            raise ValueError(
                'parameter_uncertainty_method not set for model, cannot calculate relative standard errors.'
            )


@dataclass(frozen=True)
class IIVSearchResults(ToolResults):
    pass



"""
:meta private:
"""

import math
import re
import warnings
from collections import defaultdict
from operator import add, mul
from typing import List, Literal, Set, Union

from pharmpy.deps import numpy as np
from pharmpy.deps import sympy
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs
from pharmpy.model import Assignment, Model, Parameter, Parameters, Statement, Statements

from .common import get_model_covariates
from .data import get_baselines
from .expressions import (
    depends_on,
    get_individual_parameters,
    remove_covariate_effect_from_statements,
    simplify_model,
)
from .parameters import get_thetas

EffectType = Union[Literal['lin', 'cat', 'piece_lin', 'exp', 'pow'], str]
OperationType = Literal['*', '+']


def get_covariates(model: Model) -> dict[list]:
    """Return a dictionary of all used covariates within a model

    The dictionary will have parameter name as key with a connected value as
    a list of tuple(s) with (covariate, effect type, operator)

    Parameters
    ----------
    model : Model
        Model to extract covariates from.

    Returns
    -------
        Dictionary : Dictionary of parameters and connected covariate(s)

    """
    parameters = get_individual_parameters(model)
    covariates = get_model_covariates(model)
    covariates = [c for c in covariates if model.datainfo[str(c)].type == "covariate"]

    param_w_cov = defaultdict(list)

    for p in parameters:
        for c in covariates:
            if has_covariate_effect(model, str(p), str(c)):
                param_w_cov[p].append(c)

    res = defaultdict(list)
    for param, covariates in param_w_cov.items():
        for cov in covariates:
            eff, op = get_covariate_effect(model, param, str(cov))
            res[(param, cov)].append((eff, op))

    return res


def get_covariate_effect(model: Model, symbol, covariate):
    param_expr = model.statements.before_odes.full_expression(symbol)

    # Piecewise statement can interfer with .match() of expression
    full_expr = param_expr
    function_counter = 1

    def extract_parameters_recursive(expr_list, expr, full_expr, counter):
        for a in expr.args:
            if a.is_Function:
                full_expr = full_expr.subs({a: sympy.Wild(f"f{counter}")})
                counter += 1
                if a.is_Piecewise:
                    for e, _ in a.args:
                        extract_parameters_recursive(expr_list, e, e, counter)
                else:
                    for e in a.args:
                        extract_parameters_recursive(expr_list, e, e, counter)
        expr_list.insert(0, full_expr)
        return expr_list

    param_expr_list = extract_parameters_recursive([], param_expr, full_expr, function_counter)
    param_expr_list.append(param_expr)

    eff = 'custom'
    op = '*'
    for effect in ['lin', 'cat', 'piece_lin', 'exp', 'pow']:
        template = _create_template(effect, model, covariate)
        template = template.template.expression
        wild_dict = defaultdict(list)
        for s in template.free_symbols:
            wild_symbol = sympy.Wild(str(s))
            template = template.subs({s: wild_symbol})
            if str(s).startswith("theta"):
                wild_dict["theta"].append(wild_symbol)
            elif str(s).startswith("cov"):
                wild_dict["cov"].append(wild_symbol)
            elif str(s).startswith("median"):
                wild_dict["median"].append(wild_symbol)
        rest_of_expression = sympy.Wild('reo')

        for pe in param_expr_list:
            found_match = False
            match = pe.match(rest_of_expression * template)
            if match:
                if _assert_cov_effect_match(wild_dict, match, model, covariate, effect=effect):
                    eff = effect
                    op = '*'
                    found_match = True
            if found_match:
                break
        for pe in param_expr_list:
            found_match = False
            match = pe.match(rest_of_expression + template)
            if match:
                if _assert_cov_effect_match(wild_dict, match, model, covariate, effect=effect):
                    eff = effect
                    op = '+'
                    found_match = True
            if found_match:
                break
    return eff, op


def _assert_cov_effect_match(symbols, match, model, covariate, effect):
    # TODO : Restructure covaroiate effect template matching
    if effect == "pow":
        if (
            sympy.Wild("cov") in match.keys()
            and isinstance(match[sympy.Wild("cov")], sympy.Number)
            and sympy.Wild("median") in match.keys()
            and isinstance(match[sympy.Wild("median")], sympy.Pow)
        ):
            temp = match[sympy.Wild("cov")]
            match[sympy.Wild("cov")] = match[sympy.Wild("median")]
            match[sympy.Wild("median")] = temp

    for key, values in symbols.items():
        if key == "theta":
            thetas = get_thetas(model).symbols
            if all(value in match.keys() for value in values):
                if any(match[value] not in thetas for value in values):
                    return False
        if key == "cov":
            covariate = sympy.Symbol(covariate)
            if all(value in match.keys() for value in values):
                if all(match[value] not in [covariate, 1 / covariate] for value in values):
                    return False
        if key == "median":
            if all(value in match.keys() for value in values):
                if not all(isinstance(match[value], sympy.Number) for value in values):
                    return False
    return True


def has_covariate_effect(model: Model, parameter: str, covariate: str):
    """Tests if an instance of :class:`pharmpy.model` has a given covariate
    effect.

    Parameters
    ----------
    model : Model
        Pharmpy model to check for covariate effect.
    parameter : str
        Name of parameter.
    covariate : str
        Name of covariate.

    Return
    ------
    bool
        Whether input model has a covariate effect of the input covariate on
        the input parameter.

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_covariate_effect(model, "CL", "APGR")
    False

    """
    return depends_on(model, parameter, covariate)


def remove_covariate_effect(model: Model, parameter: str, covariate: str):
    """Remove a covariate effect from an instance of :class:`pharmpy.model`.

    Parameters
    ----------
    model : Model
        Pharmpy model from which to remove the covariate effect.
    parameter : str
        Name of parameter.
    covariate : str
        Name of covariate.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_covariate_effect(model, "CL", "WGT")
    True
    >>> model = remove_covariate_effect(model, "CL", "WGT")
    >>> has_covariate_effect(model, "CL", "WGT")
    False

    """
    if not has_covariate_effect(model, parameter, covariate):
        return model
    kept_thetas, before_odes = simplify_model(
        model,
        model.statements.before_odes,
        remove_covariate_effect_from_statements(
            model, model.statements.before_odes, parameter, covariate
        ),
    )
    ode_system: List[Statement] = (
        [] if model.statements.ode_system is None else [model.statements.ode_system]
    )
    after_odes = list(model.statements.after_odes)
    statements = Statements(before_odes + ode_system + after_odes)
    kept_parameters = model.random_variables.free_symbols.union(
        kept_thetas, model.statements.after_odes.free_symbols
    )
    parameters = Parameters.create((p for p in model.parameters if p.symbol in kept_parameters))
    model = model.replace(statements=statements, parameters=parameters)

    return model.update_source()


def add_covariate_effect(
    model: Model,
    parameter: str,
    covariate: str,
    effect: EffectType,
    operation: OperationType = '*',
    allow_nested: bool = False,
):
    """Adds covariate effect to :class:`pharmpy.model`.

    The following effects have templates:

    - Linear function for continuous covariates (*lin*)
        - Function:

        .. math::

            \\text{coveff} = 1 + \\text{theta} * (\\text{cov} - \\text{median})

        - Init:  0.001
        - Upper:
            - If median of covariate equals minimum: 100,000
            - Otherwise: :math:`\\frac{1}{\\text{median} - \\text{min}}`
        - Lower:
            - If median of covariate equals maximum: -100,000
            - Otherwise: :math:`\\frac{1}{\\text{median} - \\text{max}}`
    - Linear function for categorical covariates (*cat*)
        - Function:
            - If covariate is the most common category:

            .. math::

                \\text{coveff} = 1

            - For each additional category:

            .. math::

                \\text{coveff} = 1 + \\text{theta}

        - Init: 0.001
        - Upper: 5
        - Lower: -1
    - Piecewise linear function/"hockey-stick", continuous covariates only (*piece_lin*)
        - Function:
            - If cov <= median:

            .. math::

                \\text{coveff} = 1 + \\text{theta1} * (\\text{cov} - \\text{median})

            - If cov > median:

            .. math::

                \\text{coveff} = 1 + \\text{theta2} * (\\text{cov} - \\text{median})


        - Init: 0.001
        - Upper:
            - For first state: :math:`\\frac{1}{\\text{median} - \\text{min}}`
            - Otherwise: 100,000
        - Lower:
            - For first state: -100,000
            - Otherwise: :math:`\\frac{1}{\\text{median} - \\text{max}}`
    - Exponential function, continuous covariates only (*exp*)
        - Function:

        .. math::

            \\text{coveff} = \\exp(\\text{theta} * (\\text{cov} - \\text{median}))

        - Init:
            - If lower > 0.001 or upper < 0.001: :math:`\\frac{\\text{upper} - \\text{lower}}{2}`
            - If estimated init is 0: :math:`\\frac{\\text{upper}}{2}`
            - Otherwise: 0.001
        - Upper:
            - If min - median = 0 or max - median = 0: 100
            - Otherwise:

            .. math::

                \\min(\\frac{\\log(0.01)}{\\text{min} - \\text{median}},
                \\frac{\\log(100)}{\\text{max} - \\text{median}})
        - Lower:
            - If min - median = 0 or max - median = 0: 0.01
            - Otherwise:

            .. math::

                \\max(\\frac{\\log(0.01)}{\\text{max} - \\text{median}},
                \\frac{\\log(100)}{\\text{min} - \\text{median}})

    - Power function, continuous covariates only (*pow*)
        - Function:

        .. math::

            \\text{coveff} = (\\frac{\\text{cov}}{\\text{median}})^\\text{theta}

        - Init: 0.001
        - Upper: 100,000
        - Lower: -100


    Parameters
    ----------
    model : Model
        Pharmpy model to add covariate effect to.
    parameter : str
        Name of parameter to add covariate effect to.
    covariate : str
        Name of covariate.
    effect : str
        Type of covariate effect. May be abbreviated covariate effect (see above) or custom.
    operation : str, optional
        Whether the covariate effect should be added or multiplied (default).
    allow_nested: bool, optional
        Whether to allow adding a covariate effect when one already exists for
        the input parameter-covariate pair.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = add_covariate_effect(model, "CL", "APGR", "exp")
    >>> model.statements.before_odes.full_expression("CL")
    PTVCL*WGT*exp(ETA_1)*exp(POP_CLAPGR*(APGR - 7.0))

    """
    sset = model.statements

    if not allow_nested and depends_on(model, parameter, covariate):
        warnings.warn(f'Covariate effect of {covariate} on {parameter} already exists')
        return model

    statistics = {}
    statistics['mean'] = _calculate_mean(model.dataset, covariate)
    statistics['median'] = _calculate_median(model, covariate)
    statistics['std'] = _calculate_std(model, covariate)

    covariate_effect = _create_template(effect, model, covariate)
    pset, thetas = _create_thetas(model, parameter, effect, covariate, covariate_effect.template)
    covariate_effect.apply(parameter, covariate, thetas, statistics)
    # NOTE: We hoist the statistic statements to avoid referencing variables
    # before declaring them. We also avoid duplicate statements.
    sset = [s for s in covariate_effect.statistic_statements if s not in sset] + sset

    last_existing_parameter_assignment = sset.find_assignment(parameter)
    assert last_existing_parameter_assignment is not None
    insertion_index = sset.index(last_existing_parameter_assignment) + 1

    # NOTE: We can use any assignment to the parameter since we currently only
    # use its symbol to create the new effect statement.
    effect_statement = covariate_effect.create_effect_statement(
        operation, last_existing_parameter_assignment
    )

    statements = []

    statements.append(covariate_effect.template)
    statements.append(effect_statement)

    cov_possible = {sympy.Symbol(parameter)} | {
        sympy.Symbol(f'{parameter}{col_name}') for col_name in model.datainfo.names
    }

    # NOTE: This is a heuristic that simplifies the NONMEM statements by
    # grouping multiple effect statements in a single statement.
    if last_existing_parameter_assignment.expression.args and all(
        map(cov_possible.__contains__, last_existing_parameter_assignment.expression.args)
    ):
        statements[-1] = Assignment(
            effect_statement.symbol,
            subs(
                effect_statement.expression,
                {parameter: last_existing_parameter_assignment.expression},
                simultaneous=True,
            ),
        )
        sset = sset[0 : insertion_index - 1] + sset[insertion_index:]
        insertion_index -= 1

    sset = sset[0:insertion_index] + statements + sset[insertion_index:]
    model = model.replace(parameters=pset, statements=sset)
    return model.update_source()


def natural_order(string, _nsre=re.compile(r'([0-9]+)')):
    # From https://stackoverflow.com/a/16090640
    return [int(key) if key.isdigit() else (key.lower(), key) for key in _nsre.split(string)]


def _create_thetas(model, parameter, effect, covariate, template, _ctre=re.compile(r'theta\d*')):
    """Creates theta parameters and adds to parameter set of model.

    Number of parameters depends on how many thetas have been declared."""
    new_thetas = {sym for sym in map(str, template.expression.free_symbols) if _ctre.match(sym)}
    no_of_thetas = len(new_thetas)

    pset = model.parameters

    theta_names = {}

    if no_of_thetas == 1:
        inits = _choose_param_inits(effect, model, covariate)

        theta_name = f'POP_{parameter}{covariate}'
        pset = Parameters.create(
            list(pset) + [Parameter(theta_name, inits['init'], inits['lower'], inits['upper'])]
        )
        theta_names['theta'] = theta_name
    else:
        for i, new_theta in enumerate(sorted(new_thetas, key=natural_order), 1):
            inits = _choose_param_inits(effect, model, covariate, i)

            theta_name = f'POP_{parameter}{covariate}_{i}'
            pset = Parameters.create(
                list(pset) + [Parameter(theta_name, inits['init'], inits['lower'], inits['upper'])]
            )
            theta_names[new_theta] = theta_name

    return pset, theta_names


def _count_categorical(model, covariate):
    """Gets the number of individuals that has a level of categorical covariate."""
    idcol = model.datainfo.id_column.name
    df = model.dataset.set_index(idcol)
    allcounts = df[covariate].groupby('ID').value_counts()
    allcounts.name = None  # To avoid collisions when resetting index
    counts = allcounts.reset_index().iloc[:, 1].value_counts()
    counts.sort_index(inplace=True)  # To make deterministic in case of multiple modes
    if model.dataset[covariate].isna().any():
        counts[np.nan] = 0
    return counts


def _calculate_mean(df, covariate, baselines=False):
    """Calculate mean. Can be set to use baselines, otherwise it is
    calculated first per individual, then for the group."""
    if baselines:
        return df[str(covariate)].mean()
    else:
        return df.groupby('ID')[str(covariate)].mean().mean()


def _calculate_median(model, covariate, baselines=False):
    """Calculate median. Can be set to use baselines, otherwise it is
    calculated first per individual, then for the group."""
    if baselines:
        return get_baselines(model)[str(covariate)].median()
    else:
        df = model.dataset
        return df.groupby('ID')[str(covariate)].median().median()


def _calculate_std(model, covariate, baselines=False):
    """Calculate median. Can be set to use baselines, otherwise it is
    calculated first per individual, then for the group."""
    if baselines:
        return get_baselines(model)[str(covariate)].std()
    else:
        df = model.dataset
        return df.groupby('ID')[str(covariate)].mean().std()


def _choose_param_inits(effect, model, covariate, index=None):
    """Chooses inits for parameters. If the effect is exponential, the
    bounds need to be dynamic."""
    df = model.dataset
    init_default = 0.001

    inits = {}

    cov_median = _calculate_median(model, covariate)
    cov_min = df[str(covariate)].min()
    cov_max = df[str(covariate)].max()

    lower, upper = _choose_bounds(effect, cov_median, cov_min, cov_max, index)

    if effect == 'exp':
        if lower > init_default or init_default > upper:
            init = (upper + lower) / 2
            if init == 0:
                init = upper / 5
        else:
            init = init_default
    elif effect == 'pow':
        init = init_default
    else:
        init = init_default

    inits['init'] = init
    inits['lower'] = lower
    inits['upper'] = upper

    return inits


def _choose_bounds(effect, cov_median, cov_min, cov_max, index=None):
    if effect == 'exp':
        min_diff = cov_min - cov_median
        max_diff = cov_max - cov_median

        lower_expected = 0.01
        upper_expected = 100

        if min_diff == 0 or max_diff == 0:
            return lower_expected, upper_expected
        else:
            log_base = 10
            lower = max(
                math.log(lower_expected, log_base) / max_diff,
                math.log(upper_expected, log_base) / min_diff,
            )
            upper = min(
                math.log(lower_expected, log_base) / min_diff,
                math.log(upper_expected, log_base) / max_diff,
            )
    elif effect == 'lin':
        if cov_median == cov_min:
            upper = 100000
        else:
            upper = 1 / (cov_median - cov_min)
        if cov_median == cov_max:
            lower = -100000
        else:
            lower = 1 / (cov_median - cov_max)
    elif effect == 'piece_lin':
        if cov_median == cov_min or cov_median == cov_max:
            raise Exception(
                'Median cannot be same as min or max, cannot use '
                'piecewise-linear parameterization.'
            )
        if index == 0:
            lower = -100000
            upper = 1 / (cov_median - cov_min)
        else:
            lower = 1 / (cov_median - cov_max)
            upper = 100000
    elif effect == 'pow':
        lower = -100
        upper = 100000
    elif effect == 'cat':
        lower = -1
        upper = 5
    else:
        lower = -100000
        upper = 100000
    return round(lower, 4), round(upper, 4)


def _create_template(effect, model, covariate):
    """Creates Covariate class objects with effect template."""
    if effect == 'lin':
        return CovariateEffect.linear()
    elif effect == 'cat':
        counts = _count_categorical(model, covariate)
        return CovariateEffect.categorical(counts)
    elif effect == 'piece_lin':
        return CovariateEffect.piecewise_linear()
    elif effect == 'exp':
        return CovariateEffect.exponential()
    elif effect == 'pow':
        return CovariateEffect.power()
    else:
        symbol = sympy.Symbol('symbol')
        expression = parse_expr(effect)
        return CovariateEffect(Assignment(symbol, expression))


class CovariateEffect:
    """
    Covariate effect consisting of new assignments.

    Attributes
    ----------
    template
        Assignment based on covariate effect
    statistic_statements
        Dict with mean, median and standard deviation

    :meta private:

    """

    def __init__(self, template):
        self.template = template
        self.statistic_statements = []

    def apply(self, parameter, covariate, thetas, statistics):
        effect_name = f'{parameter}{covariate}'
        self.template = Assignment(
            sympy.Symbol(effect_name),
            subs(
                subs(
                    self.template.expression,
                    thetas,
                ),
                {'cov': covariate},
                simultaneous=True,
            ),
        )

        template_str = [str(symbol) for symbol in self.template.free_symbols]

        if 'mean' in template_str:
            self.template = self.template.subs({'mean': f'{covariate}_MEAN'})
            s = Assignment(sympy.Symbol(f'{covariate}_MEAN'), sympy.Float(statistics['mean'], 6))
            self.statistic_statements.append(s)
        if 'median' in template_str:
            self.template = self.template.subs({'median': f'{covariate}_MEDIAN'})
            s = Assignment(
                sympy.Symbol(f'{covariate}_MEDIAN'), sympy.Float(statistics['median'], 6)
            )
            self.statistic_statements.append(s)
        if 'std' in template_str:
            self.template = self.template.subs({'std': f'{covariate}_STD'})
            s = Assignment(sympy.Symbol(f'{covariate}_STD'), sympy.Float(statistics['std'], 6))
            self.statistic_statements.append(s)

    def create_effect_statement(self, operation_str, statement_original):
        """Creates statement for addition or multiplication of covariate
        to parameter, e.g. (if parameter is CL and covariate is WGT):

            CL = CLWGT + TVCL*EXP(ETA(1))"""
        operation = self._get_operation(operation_str)

        symbol = statement_original.symbol
        expression = statement_original.symbol

        statement_new = Assignment(symbol, operation(expression, self.template.symbol))

        return statement_new

    @staticmethod
    def _get_operation(operation_str):
        """Gets sympy operation based on string"""
        if operation_str == '*':
            return mul
        elif operation_str == '+':
            return add

        raise NotImplementedError(f'Can only handle + or *, got {operation_str}.')

    @classmethod
    def linear(cls):
        """Linear continuous template (for continuous covariates)."""
        symbol = sympy.Symbol('symbol')
        expression = 1 + sympy.Symbol('theta') * (sympy.Symbol('cov') - sympy.Symbol('median'))
        template = Assignment(symbol, expression)

        return cls(template)

    @classmethod
    def categorical(cls, counts):
        """Linear categorical template (for categorical covariates)."""
        symbol = sympy.Symbol('symbol')
        most_common = counts.idxmax()
        categories = list(counts.index)

        values = [1]
        conditions = [sympy.Eq(sympy.Symbol('cov'), most_common)]

        for i, cat in enumerate(categories, 1):
            if cat != most_common:
                if np.isnan(cat):
                    conditions += [sympy.Eq(sympy.Symbol('cov'), sympy.Symbol('NaN'))]
                    values += [1]
                else:
                    conditions += [sympy.Eq(sympy.Symbol('cov'), cat)]
                    if len(categories) == 2:
                        values += [1 + sympy.Symbol('theta')]
                    else:
                        values += [1 + sympy.Symbol(f'theta{i}')]

        expression = sympy.Piecewise(*zip(values, conditions))

        template = Assignment(symbol, expression)

        return cls(template)

    @classmethod
    def piecewise_linear(cls):
        """Piecewise linear ("hockey-stick") template (for continuous
        covariates)."""
        symbol = sympy.Symbol('symbol')
        values = [
            1 + sympy.Symbol('theta1') * (sympy.Symbol('cov') - sympy.Symbol('median')),
            1 + sympy.Symbol('theta2') * (sympy.Symbol('cov') - sympy.Symbol('median')),
        ]
        conditions = [
            sympy.Le(sympy.Symbol('cov'), sympy.Symbol('median')),
            sympy.Gt(sympy.Symbol('cov'), sympy.Symbol('median')),
        ]
        expression = sympy.Piecewise((values[0], conditions[0]), (values[1], conditions[1]))

        template = Assignment(symbol, expression)

        return cls(template)

    @classmethod
    def exponential(cls):
        """Exponential template (for continuous covariates)."""
        symbol = sympy.Symbol('symbol')
        expression = sympy.exp(
            sympy.Symbol('theta') * (sympy.Symbol('cov') - sympy.Symbol('median'))
        )
        template = Assignment(symbol, expression)

        return cls(template)

    @classmethod
    def power(cls):
        """Power template (for continuous covariates)."""
        symbol = sympy.Symbol('symbol')
        expression = (sympy.Symbol('cov') / sympy.Symbol('median')) ** sympy.Symbol('theta')
        template = Assignment(symbol, expression)

        return cls(template)

    def __str__(self):
        """String representation of class."""
        return str(self.template)


def get_covariates_allowed_in_covariate_effect(model: Model) -> Set[str]:
    try:
        di_covariate = model.datainfo.typeix['covariate'].names
    except IndexError:
        di_covariate = []

    try:
        di_admid = model.datainfo.typeix['admid'].names
    except IndexError:
        di_admid = []

    try:
        di_unknown = model.datainfo.typeix['unknown'].names
    except IndexError:
        di_unknown = []

    return set(di_covariate).union(di_unknown, di_admid)



from __future__ import annotations

import importlib
import inspect
import re
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, List, Mapping, Optional, Sequence, Tuple, Union, get_type_hints

import pharmpy
import pharmpy.tools.modelfit
import pharmpy.workflows.results
from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.internals.fs.path import normalize_user_given_path
from pharmpy.model import Model
from pharmpy.modeling import (
    calculate_aic,
    calculate_bic,
    check_high_correlations,
    check_parameters_near_bounds,
    get_omegas,
    get_sigmas,
    get_thetas,
    read_model,
)
from pharmpy.modeling.lrt import degrees_of_freedom as lrt_df
from pharmpy.modeling.lrt import test as lrt_test
from pharmpy.tools.psn_helpers import create_results as psn_create_results
from pharmpy.workflows import Results, Workflow, execute_workflow, split_common_options
from pharmpy.workflows.model_database import LocalModelDirectoryDatabase, ModelDatabase
from pharmpy.workflows.results import ModelfitResults, mfr
from pharmpy.workflows.tool_database import ToolDatabase

from .external import parse_modelfit_results


def fit(
    model_or_models: Union[Model, List[Model]],
    tool: Optional[str] = None,
    path: Optional[Union[Path, str]] = None,
) -> Union[ModelfitResults, List[ModelfitResults]]:
    """Fit models.

    Parameters
    ----------
    model_or_models : Model | list[Model]
        List of models or one single model
    tool : str
        Estimation tool to use. None to use default
    path :  Path | str
        Path to fit directory

    Return
    ------
    ModelfitResults | list[ModelfitResults]
        ModelfitResults for the model or models

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> from pharmpy.tools import fit
    >>> model = load_example_model("pheno")      # doctest: +SKIP
    >>> results = fit(model)      # doctest: +SKIP

    See also
    --------
    run_tool

    """
    single, models = (
        (True, [model_or_models])
        if isinstance(model_or_models, Model)
        else (False, model_or_models)
    )

    kept = []
    # Do not fit model if already fit
    for model in models:
        try:
            # FIXME: model.database should be removed
            db_model = retrieve_models(model.database, model.name)[0]
        except (KeyError, AttributeError):
            db_model = None
        if (
            db_model
            and db_model.modelfit_results is not None
            and db_model == model
            and model.has_same_dataset_as(db_model)
        ):
            model = model.replace(modelfit_results=db_model.modelfit_results)
        else:
            kept.append(model)

    if kept:
        modelfit_results = run_tool('modelfit', kept, tool=tool, path=path)

    return modelfit_results if single else list(modelfit_results)


def create_results(path: Union[str, Path], **kwargs) -> Results:
    """Create/recalculate results object given path to run directory

    Parameters
    ----------
    path : str, Path
        Path to run directory
    kwargs
        Arguments to pass to tool specific create results function

    Returns
    -------
    Results
        Results object for tool

    Examples
    --------
    >>> from pharmpy.tools import create_results
    >>> res = create_results("frem_dir1")   # doctest: +SKIP

    See also
    --------
    read_results

    """
    path = normalize_user_given_path(path)
    res = psn_create_results(path, **kwargs)
    return res


def read_results(path: Union[str, Path]) -> Results:
    """Read results object from file

    Parameters
    ----------
    path : str, Path
        Path to results file

    Return
    ------
    Results
        Results object for tool

    Examples
    --------
    >>> from pharmpy.tools import read_results
    >>> res = read_results("results.json")     # doctest: +SKIP

    See also
    --------
    create_results

    """
    path = normalize_user_given_path(path)
    res = pharmpy.workflows.results.read_results(path)
    return res


def run_tool(name: str, *args, **kwargs) -> Union[Model, List[Model], Tuple[Model], Results]:
    """Run tool workflow

    Parameters
    ----------
    name : str
        Name of tool to run
    args
        Arguments to pass to tool
    kwargs
        Arguments to pass to tool

    Return
    ------
    Results
        Results object for tool

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> from pharmpy.tools import run_tool # doctest: +SKIP
    >>> res = run_tool("ruvsearch", model)   # doctest: +SKIP

    """
    # NOTE: The implementation of run_tool is split into those two functions to
    # allow for individual testing and mocking.
    tool = import_tool(name)
    return run_tool_with_name(name, tool, args, kwargs)


def import_tool(name: str):
    return importlib.import_module(f'pharmpy.tools.{name}')


def run_tool_with_name(
    name: str, tool, args: Sequence, kwargs: Mapping[str, Any]
) -> Union[Model, List[Model], Tuple[Model], Results]:
    # FIXME: Workaround until ModelfitResults is disentangled with
    #  Model object
    if 'model' in kwargs and 'results' in kwargs:
        model = kwargs['model']
        res = kwargs['results']
        if isinstance(model, Model) and isinstance(res, ModelfitResults):
            model = model.replace(modelfit_results=res)
            kwargs['model'] = model

    common_options, tool_options = split_common_options(kwargs)

    create_workflow = tool.create_workflow

    dispatcher, tool_database = _get_run_setup(common_options, name)

    tool_params = inspect.signature(create_workflow).parameters
    tool_param_types = get_type_hints(create_workflow)

    tool_metadata = _create_metadata(
        database=tool_database,
        dispatcher=dispatcher,
        tool_name=name,
        tool_params=tool_params,
        tool_param_types=tool_param_types,
        args=args,
        tool_options=tool_options,
        common_options=common_options,
    )

    tool_database.store_metadata(tool_metadata)

    if validate_input := getattr(tool, 'validate_input', None):
        validate_input(*args, **tool_options)

    wf: Workflow = create_workflow(*args, **tool_options)
    assert wf.name == name

    res = execute_workflow(wf, dispatcher=dispatcher, database=tool_database)
    assert name == 'modelfit' or isinstance(res, Results)

    tool_metadata = _update_metadata(tool_metadata, res)
    tool_database.store_metadata(tool_metadata)

    return res


def _create_metadata(
    database: ToolDatabase,
    dispatcher,
    tool_name: str,
    tool_params,
    tool_param_types,
    args: Sequence,
    tool_options: Mapping[str, Any],
    common_options: Mapping[str, Any],
):
    tool_metadata = _create_metadata_tool(
        database, tool_name, tool_params, tool_param_types, args, tool_options
    )
    setup_metadata = _create_metadata_common(database, dispatcher, tool_name, common_options)
    tool_metadata['common_options'] = setup_metadata

    return tool_metadata


def _update_metadata(tool_metadata, res):
    # FIXME: Make metadata immutable
    tool_metadata['stats']['end_time'] = _now()
    return tool_metadata


def resume_tool(path: str):
    """Resume tool workflow from tool database path

    Parameters
    ----------
    path : str
        The path to the tool database

    Return
    ------
    Results
        Results object for tool

    Examples
    --------
    >>> from pharmpy.modeling import * # doctest: +SKIP
    >>> res = resume_tool("resmod_dir1") # doctest: +SKIP

    """

    dispatcher, tool_database = _get_run_setup_from_metadata(path)

    tool_metadata = tool_database.read_metadata()
    tool_name = tool_metadata['tool_name']

    tool = importlib.import_module(f'pharmpy.tools.{tool_name}')

    create_workflow = tool.create_workflow

    tool_params = inspect.signature(create_workflow).parameters
    tool_param_types = get_type_hints(create_workflow)

    tool_options = _parse_tool_options_from_json_metadata(
        tool_metadata, tool_params, tool_param_types, tool_database
    )

    args, kwargs = _parse_args_kwargs_from_tool_options(tool_params, tool_options)

    if validate_input := getattr(tool, 'validate_input', None):
        validate_input(*args, **kwargs)

    wf: Workflow = create_workflow(*args, **kwargs)
    assert wf.name == tool_name

    res = execute_workflow(wf, dispatcher=dispatcher, database=tool_database)
    assert tool_name == 'modelfit' or isinstance(res, Results)

    tool_metadata = _update_metadata(tool_metadata, res)
    tool_database.store_metadata(tool_metadata)

    return res


def _parse_tool_options_from_json_metadata(
    tool_metadata,
    tool_params,
    tool_param_types,
    tool_database,
):
    tool_options = tool_metadata['tool_options']
    # NOTE: Load models to memory
    for model_key in _input_model_param_keys(tool_params, tool_param_types):
        model_metadata = tool_options.get(model_key)
        if model_metadata is None:
            raise ValueError(
                f'Cannot resume run because model argument "{model_key}" cannot be restored.'
            )

        assert model_metadata['__class__'] == 'Model'
        model_name = model_metadata['arg_name']
        db_name = model_metadata['db_name']

        db: ModelDatabase = tool_database.model_database
        try:
            model = db.retrieve_model(db_name)
            model = model.replace(name=model_name)
            res = db.retrieve_modelfit_results(db_name)
            model = model.replace(modelfit_results=res)
        except KeyError:
            raise ValueError(
                f'Cannot resume run because model argument "{model_key}" ({model_name}) cannot be restored.'
            )
        tool_options = tool_options.copy()
        tool_options[model_key] = model

    # NOTE: Load results to memory
    for results_key in _results_param_keys(tool_params, tool_param_types):
        results_json = tool_options.get(results_key)
        if results_json is not None:
            tool_options = tool_options.copy()
            tool_options[results_key] = pharmpy.workflows.results.read_results(results_json)

    return tool_options


def _parse_args_kwargs_from_tool_options(tool_params, tool_options):
    args = []
    kwargs = {}
    for p in tool_params.values():
        # Positional args
        if p.default == p.empty:
            args.append(tool_options[p.name])
        # Named args
        else:
            if p.name in tool_options.keys():
                kwargs[p.name] = tool_options[p.name]

    return args, kwargs


def _create_metadata_tool(
    database: ToolDatabase,
    tool_name: str,
    tool_params,
    tool_param_types,
    args: Sequence,
    kwargs: Mapping[str, Any],
):
    # FIXME: Add config file dump, estimation tool etc.
    tool_metadata = {
        'pharmpy_version': pharmpy.__version__,
        'tool_name': tool_name,
        'stats': {'start_time': _now()},
        'tool_options': {},
    }

    for i, p in enumerate(tool_params.values()):
        try:
            name, value = p.name, args[i]
        except IndexError:
            try:
                name, value = p.name, kwargs[p.name]
            except KeyError:
                if p.default == p.empty:
                    # Positional args
                    raise ValueError(f'{tool_name}: \'{p.name}\' was not set')
                else:
                    # Named args
                    name, value = p.name, p.default

        tool_metadata['tool_options'][name] = value

    if tool_name != 'modelfit':
        db = database.model_database
        for key, arg_name, db_name in _store_input_models(
            db, tool_params, tool_param_types, args, kwargs
        ):
            tool_metadata['tool_options'][key] = {
                '__class__': 'Model',
                'arg_name': arg_name,
                'db_name': db_name,
            }

    return tool_metadata


def _create_metadata_common(
    database: ToolDatabase, dispatcher, toolname: Optional[str], common_options: Mapping[str, Any]
):
    setup_metadata = {}
    setup_metadata['dispatcher'] = dispatcher.__name__
    # FIXME: Naming of workflows/tools should be consistent (db and input name of tool)
    setup_metadata['database'] = {
        'class': type(database).__name__,
        'toolname': toolname,
        'path': str(database.path),
    }
    for key, value in common_options.items():
        if key not in setup_metadata.keys():
            if isinstance(value, Path):
                value = str(value)
            setup_metadata[str(key)] = value

    return setup_metadata


def _store_input_models(
    db: ModelDatabase, params, types, args: Sequence, kwargs: Mapping[str, Any]
):
    for param_key, model in _input_models(params, types, args, kwargs):
        input_model_name = f'input_{param_key}'
        _store_input_model(db, model, input_model_name)
        yield param_key, model.name, input_model_name


def _filter_params(kind, params, types):
    for i, param_key in enumerate(params):
        param = params[param_key]
        param_type = types.get(param_key)
        if param_type in (kind, Optional[kind]):
            # NOTE: We do not handle *{param_key}, or **{param_key}
            assert param.kind != param.VAR_POSITIONAL
            assert param.kind != param.VAR_KEYWORD
            yield i, param_key


def _input_model_param_keys(params, types):
    for _, param_key in _filter_params(Model, params, types):
        yield param_key


def _results_param_keys(params, types):
    for _, param_key in _filter_params(ModelfitResults, params, types):
        yield param_key


def _input_models(params, types, args: Sequence, kwargs: Mapping[str, Any]):
    for i, param_key in _filter_params(Model, params, types):
        model = args[i] if i < len(args) else kwargs.get(param_key)
        if model is None:
            continue
        yield param_key, model


def _store_input_model(db: ModelDatabase, model: Model, name: str):
    model_copy = model.replace(name=name)
    with db.transaction(model_copy) as txn:
        txn.store_model()
        txn.store_modelfit_results()


def _now():
    return datetime.now().astimezone().isoformat()


def _get_run_setup(common_options, toolname) -> Tuple[Any, ToolDatabase]:
    try:
        dispatcher = common_options['dispatcher']
    except KeyError:
        from pharmpy.workflows import default_dispatcher

        dispatcher = default_dispatcher

    try:
        database = common_options['database']
    except KeyError:
        from pharmpy.workflows import default_tool_database

        if 'path' in common_options.keys():
            path = common_options['path']
        else:
            path = None
        database = default_tool_database(
            toolname=toolname, path=path, exist_ok=common_options.get('resume', False)
        )  # TODO: database -> tool_database

    return dispatcher, database


def retrieve_models(
    source: Union[str, Path, Results, ToolDatabase, ModelDatabase],
    names: Optional[List[str]] = None,
) -> List[Model]:
    """Retrieve models after a tool run

    Any models created and run by the tool can be
    retrieved.

    Parameters
    ----------
    source : str, Path, Results, ToolDatabase, ModelDatabase
        Source where to find models. Can be a path (as str or Path), a results object, or a
        ToolDatabase/ModelDatabase
    names : list
        List of names of the models to retrieve or None for all

    Return
    ------
    list
        List of retrieved model objects

    Examples
    --------
    >>> from pharmpy.tools import retrieve_models
    >>> tooldir_path = 'path/to/tool/directory'
    >>> models = retrieve_models(tooldir_path, names=['run1'])      # doctest: +SKIP

    See also
    --------
    retrieve_final_model

    """
    if isinstance(source, Path) or isinstance(source, str):
        path = Path(source)
        # FIXME: Should be using metadata to know how to init databases
        db = LocalModelDirectoryDatabase(path / 'models')
    elif isinstance(source, Results):
        try:
            db_tool = getattr(source, 'tool_database')
            db = db_tool.model_database
        except AttributeError:
            raise ValueError(
                f'Results type \'{source.__class__.__name__}\' does not serialize tool database'
            )
    elif isinstance(source, ToolDatabase):
        db = source.model_database
    elif isinstance(source, ModelDatabase):
        db = source
    else:
        raise NotImplementedError(f'Not implemented for type \'{type(source)}\'')
    names_all: List[str] = db.list_models()
    if names is None:
        names = names_all
    diff = set(names).difference(names_all)
    if diff:
        raise ValueError(f'Models {diff} not in database')
    models = [db.retrieve_model(name) for name in names]
    return models


def retrieve_final_model(res: Results) -> Model:
    """Retrieve final model from a result object

    Parameters
    ----------
    res : Results
        A results object

    Return
    ------
    Model
        Reference to final model

    Examples
    --------
    >>> from pharmpy.tools import read_results, retrieve_final_model
    >>> res = read_results("results.json")     # doctest: +SKIP
    >>> model = retrieve_final_model(res)      # doctest: +SKIP

    See also
    --------
    retrieve_models

    """
    try:
        final_model = getattr(res, 'final_model')
    except AttributeError:
        raise ValueError('Attribute \'final_model\' is missing from results object')

    if final_model is None:
        raise ValueError('Attribute \'final_model\' is None')

    return retrieve_models(res, names=[final_model.name])[0]


def print_fit_summary(model: Model):
    """Print a summary of the model fit

    Parameters
    ----------
    model : Model
        Pharmpy model object
    """

    def bool_ok_error(x):
        return "OK" if x else "ERROR"

    def bool_yes_no(x):
        return "YES" if x else "NO"

    def print_header(text, first=False):
        if not first:
            print()
        print(text)
        print("-" * len(text))

    def print_fmt(text, result):
        print(f"{text:33} {result}")

    res = mfr(model)

    print_header("Parameter estimation status", first=True)
    print_fmt("Minimization successful", bool_ok_error(res.minimization_successful))
    print_fmt("No rounding errors", bool_ok_error(res.termination_cause != 'rounding_errors'))
    ofv = res.ofv
    assert ofv is not None
    print_fmt("Objective function value", round(ofv, 1))

    print_header("Parameter uncertainty status")
    cov_run = model.estimation_steps[-1].cov
    print_fmt("Covariance step run", bool_yes_no(cov_run))

    if cov_run:
        condno = round(np.linalg.cond(res.correlation_matrix), 1)
        print_fmt("Condition number", condno)
        print_fmt("Condition number < 1000", bool_ok_error(condno < 1000))
        cor = res.correlation_matrix
        assert cor is not None
        hicorr = check_high_correlations(model, cor)
        print_fmt("No correlations arger than 0.9", bool_ok_error(hicorr.empty))

    print_header("Parameter estimates")
    pe = res.parameter_estimates
    if cov_run:
        se = res.standard_errors
        assert se is not None
        rse = se / pe
        rse.name = 'RSE'
        df = pd.concat([pe, se, rse], axis=1)
    else:
        df = pd.concat([pe], axis=1)
    print(df)


def write_results(results: Results, path: Union[str, Path], lzma: bool = False, csv: bool = False):
    """Write results object to json (or csv) file

    Note that the csv-file cannot be read into a results object again.

    Parameters
    ----------
    results : Results
        Pharmpy results object
    path : Path
        Path to results file
    lzma : bool
        True for lzma compression. Not applicable to csv file
    csv : bool
        Save as csv file
    """
    path = normalize_user_given_path(path)
    if csv:
        results.to_csv(path)
    else:
        results.to_json(path, lzma=lzma)


def summarize_errors(results: Union[ModelfitResults, List[ModelfitResults]]) -> pd.DataFrame:
    """Summarize errors and warnings from one or multiple model runs.

    Summarize the errors and warnings found after running the model/models.

    Parameters
    ----------
    results : list, ModelfitResults
        List of ModelfitResults or single ModelfitResults

    Return
    ------
    pd.DataFrame
        A DataFrame of errors with model name, category (error or warning), and an int as index,
        an empty DataFrame if there were no errors or warnings found.

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> from pharmpy.tools import summarize_errors
    >>> model = load_example_model("pheno")
    >>> summarize_errors(model)      # doctest: +SKIP
    """
    # FIXME: Have example with errors
    if isinstance(results, ModelfitResults):
        results = [results]

    idcs, rows = [], []

    for res in results:
        if res is not None and len(res.log) > 0:
            for i, entry in enumerate(res.log):
                idcs.append((res.name, entry.category, i))
                rows.append([entry.time, entry.message])

    index_names = ['model', 'category', 'error_no']
    col_names = ['time', 'message']
    index = pd.MultiIndex.from_tuples(idcs, names=index_names)

    if rows:
        df = pd.DataFrame(rows, columns=col_names, index=index)
    else:
        df = pd.DataFrame(columns=col_names, index=index)

    return df.sort_index()


def rank_models(
    base_model: Model,
    base_model_res: ModelfitResults,
    models: List[Model],
    models_res: List[ModelfitResults],
    parent_dict: Optional[Union[dict[str, str], dict[Model, Model]]] = None,
    strictness: Optional[str] = "minimization_successful",
    rank_type: str = 'ofv',
    cutoff: Optional[float] = None,
    bic_type: str = 'mixed',
    **kwargs,
) -> pd.DataFrame:
    """Ranks a list of models

    Ranks a list of models with a given ranking function

    Parameters
    ----------
    base_model : Model
        Base model to compare to
    base_model_res : ModelfitResults
        Results of base model
    models : list
        List of models
    models_res : list
        List of modelfit results
    parent_dict : dict
        Dict where key is child and value is parent. Only relevant for LRT, if None base will be set as parent
    strictness : str or None
        Strictness criteria that are allowed for ranking. Default is "minimization_successful".
    rank_type : str
        Name of ranking type. Available options are 'ofv', 'aic', 'bic', 'lrt' (OFV with LRT)
    cutoff : float or None
        Value to use as cutoff. If using LRT, cutoff denotes p-value. Default is None
    bic_type : str
        Type of BIC to calculate. Default is the mixed effects.
    kwargs
        Arguments to pass to calculate BIC (such as `mult_test_p` and `mult_test_p`)

    Return
    ------
    pd.DataFrame
        DataFrame of the ranked models

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> from pharmpy.tools import rank_models
    >>> model_1 = load_example_model("pheno")
    >>> model_2 = load_example_model("pheno_linear")
    >>> rank_models(model_1, [model_2],
    ...             rank_type='lrt') # doctest: +SKIP
    """
    if len(models) != len(models_res):
        raise ValueError('Different length of `models` and `models_res`')
    if rank_type == 'lrt' and not parent_dict:
        parent_dict = {model.name: base_model.name for model in models}
    if parent_dict and not isinstance(list(parent_dict.keys())[0], str):
        parent_dict = {child.name: parent.name for child, parent in parent_dict.items()}

    models_all = [base_model] + models
    res_all = [base_model_res] + models_res

    rank_values, delta_values = {}, {}
    models_to_rank = []

    ref_value = _get_rankval(base_model, base_model_res, strictness, rank_type, bic_type, **kwargs)
    model_dict = {model.name: (model, res) for model, res in zip(models_all, res_all)}

    # Filter on strictness
    for model, res in zip(models_all, res_all):
        # Exclude OFV etc. if model was not successful
        rank_value = _get_rankval(model, res, strictness, rank_type, bic_type, **kwargs)
        if np.isnan(rank_value):
            continue
        if model.name == base_model.name:
            pass
        elif rank_type == 'lrt':
            parent_model, parent_res = model_dict[parent_dict[model.name]]
            if cutoff is None:
                co = 0.05 if lrt_df(parent_model, model) >= 0 else 0.01
            elif isinstance(cutoff, tuple):
                co = cutoff[0] if lrt_df(parent_model, model) >= 0 else cutoff[1]
            else:
                assert isinstance(cutoff, (float, int))
                co = cutoff
            parent_ofv = np.nan if (mfr := parent_res) is None else mfr.ofv
            model_ofv = np.nan if (mfr := res) is None else mfr.ofv
            if not lrt_test(parent_model, model, parent_ofv, model_ofv, co):
                continue
        elif cutoff is not None:
            if ref_value - rank_value <= cutoff:
                continue

        # Add ranking value and model
        rank_values[model.name] = rank_value
        delta_values[model.name] = ref_value - rank_value
        models_to_rank.append(model)

    # Sort
    def _get_delta(model):
        if np.isnan(ref_value):
            return -rank_values[model.name]
        return delta_values[model.name]

    models_sorted = sorted(models_to_rank, key=_get_delta, reverse=True)

    # Create rank for models, if two have the same value they will have the same rank
    ranking = {}
    rank, count, prev = 0, 0, None
    for model in models_sorted:
        count += 1
        value = _get_delta(model)
        if value != prev:
            rank += count
            prev = value
            count = 0
        ranking[model.name] = rank

    rows = {}
    for model in models_all:
        delta, rank_value, rank = np.nan, np.nan, np.nan
        if model.name in ranking.keys():
            rank = ranking[model.name]
        if model.name in rank_values.keys():
            rank_value = rank_values[model.name]
        if model.name in delta_values.keys():
            delta = delta_values[model.name]

        rows[model.name] = (delta, rank_value, rank)

    if rank_type == 'lrt':
        rank_type_name = 'ofv'
    else:
        rank_type_name = rank_type

    index = pd.Index(rows.keys(), name='model')
    df = pd.DataFrame(
        rows.values(), index=index, columns=[f'd{rank_type_name}', f'{rank_type_name}', 'rank']
    )

    if np.isnan(ref_value):
        return df.sort_values(by=[f'{rank_type_name}'])
    else:
        return df.sort_values(by=[f'd{rank_type_name}'], ascending=False)


class ArrayEvaluator:
    def __init__(self, x):
        self.x = x

    def __lt__(self, value):
        return all(e < value for e in self.x)

    def __eq__(self, value):
        return all(e == value for e in self.x)

    def __le__(self, value):
        return all(e <= value for e in self.x)

    def __ge__(self, value):
        return all(e >= value for e in self.x)

    def __gt__(self, value):
        return all(e > value for e in self.x)


def is_strictness_fulfilled(
    res: ModelfitResults,
    model: Model,
    statement: str,
) -> bool:
    """Takes a ModelfitResults object and a statement as input and returns True/False
    if the evaluation of the statement is True/False.

    Parameters
    ----------
    results : ModelfitResults
        ModelfitResults object
    model : Model
        Model for parameter specific strictness.
    statement : str
        A statement containing the strictness criteria

    Return
    ------
    bool
        A bool indicating whether the strictness criteria are fulfilled or not.

    Examples
    --------
    >>> from pharmpy.tools import *
    >>> from pharmpy.modeling import *
    >>> res = load_example_modelfit_results('pheno')
    >>> model = load_example_model('pheno')
    >>> is_strictness_fulfilled(res, model, "minimization_successful or rounding_errors")
    True
    """
    if res is None or np.isnan(res.ofv):
        return False
    if statement is not None:
        statement = statement.lower()
        allowed_args = [
            'minimization_successful',
            'rounding_errors',
            'sigdigs',
            'maxevals_exceeded',
            'rse',
            'rse_theta',
            'rse_omega',
            'rse_sigma',
            'condition_number',
            'final_zero_gradient',
            'final_zero_gradient_theta',
            'final_zero_gradient_omega',
            'final_zero_gradient_sigma',
            'estimate_near_boundary',
            'estimate_near_boundary_theta',
            'estimate_near_boundary_omega',
            'estimate_near_boundary_sigma',
        ]
        unwanted_args = ['and', 'or', 'not']
        find_all_words = re.findall(r'[^\d\W]+', statement)
        args_in_statement = [w for w in find_all_words if w not in unwanted_args]
        find_all_non_allowed_operators = re.findall(r"[^\w\s\.\<\>\=\!\(\)]", statement)
        if len(find_all_non_allowed_operators) > 0:
            raise ValueError(
                f"Unallowed operators found: {', '.join(find_all_non_allowed_operators)}"
            )

        # Check that only allowed arguments are in the statement
        if not all(map(lambda x: x in allowed_args, args_in_statement)):
            raise ValueError(
                f'Some expressions were not correct. Valid arguments are: {allowed_args}'
            )
        else:
            minimization_successful = res.minimization_successful  # noqa
            rounding_errors = res.termination_cause == "rounding_errors"  # noqa
            maxevals_exceeded = res.termination_cause == "maxevals_exceeded"  # noqa
            sigdigs = ArrayEvaluator([res.significant_digits])  # noqa
            final_zero_gradient = 'final_zero_gradient' in res.warnings  # noqa
            estimate_near_boundary = 'estimate_near_boundary' in res.warnings  # noqa
            if 'condition_number' in args_in_statement:
                if res.covariance_matrix is not None:
                    condition_number = ArrayEvaluator(  # noqa
                        [np.linalg.cond(res.covariance_matrix)]
                    )
                else:
                    raise ValueError("Could not calculate condition_number.")
            if "rse" in args_in_statement:
                if res.relative_standard_errors is not None:
                    rse = ArrayEvaluator(res.relative_standard_errors)  # noqa
                else:
                    raise ValueError("Could not calculate relative standard error.")

            if (
                'rse_theta' in args_in_statement
                or 'rse_omega' in args_in_statement
                or 'rse_sigma' in args_in_statement
            ):
                rse = res.relative_standard_errors
                rse_theta = ArrayEvaluator(rse[rse.index.isin(get_thetas(model).names)])  # noqa
                rse_omega = ArrayEvaluator(rse[rse.index.isin(get_omegas(model).names)])  # noqa
                rse_sigma = ArrayEvaluator(rse[rse.index.isin(get_sigmas(model).names)])  # noqa
            if (
                'final_zero_gradient_theta' in args_in_statement
                or 'final_zero_gradient_omega' in args_in_statement
                or 'final_zero_gradient_sigma' in args_in_statement
            ):
                grd = res.gradients
                final_zero_gradient_theta = (  # noqa
                    grd[grd.index.isin(get_thetas(model).names)] == 0
                ).any() or grd[grd.index.isin(get_thetas(model).names)].isnull().any()
                final_zero_gradient_omega = (  # noqa
                    grd[grd.index.isin(get_omegas(model).names)] == 0
                ).any() or grd[grd.index.isin(get_thetas(model).names)].isnull().any()
                final_zero_gradient_sigma = (  # noqa
                    grd[grd.index.isin(get_sigmas(model).names)] == 0
                ).any() or grd[grd.index.isin(get_thetas(model).names)].isnull().any()
            if (
                'estimate_near_boundary' in args_in_statement
                or 'estimate_near_boundary_theta' in args_in_statement
                or 'estimate_near_boundary_omega' in args_in_statement
                or 'estimate_near_boundary_sigma' in args_in_statement
            ):
                ests = res.parameter_estimates
                estimate_near_boundary = check_parameters_near_bounds(model, ests).any()  # noqa
                estimate_near_boundary_theta = check_parameters_near_bounds(  # noqa
                    model, ests[ests.index.isin(get_thetas(model).names)]
                ).any()
                estimate_near_boundary_omega = check_parameters_near_bounds(  # noqa
                    model, ests[ests.index.isin(get_omegas(model).names)]
                ).any()
                estimate_near_boundary_sigma = check_parameters_near_bounds(  # noqa
                    model, ests[ests.index.isin(get_sigmas(model).names)]
                ).any()

        return eval(statement)
    else:
        return True


def _get_rankval(model, res, strictness, rank_type, bic_type, **kwargs):
    if not is_strictness_fulfilled(res, model, strictness):
        return np.nan
    if rank_type in ['ofv', 'lrt']:
        return res.ofv
    elif rank_type == 'aic':
        return calculate_aic(model, res.ofv)
    elif rank_type == 'bic':
        return calculate_bic(model, res.ofv, bic_type, **kwargs)
    else:
        raise ValueError('Unknown rank_type: must be ofv, lrt, aic, or bic')


def summarize_modelfit_results(
    results: Union[ModelfitResults, List[ModelfitResults]],
    include_all_estimation_steps: bool = False,
) -> pd.DataFrame:
    """Summarize results of model runs

    Summarize different results after fitting a model, includes runtime, ofv,
    and parameter estimates (with errors). If include_all_estimation_steps is False,
    only the last estimation step will be included (note that in that case, the
    minimization_successful value will be referring to the last estimation step, if
    last step is evaluation it will go backwards until it finds an estimation step
    that wasn't an evaluation).

    Parameters
    ----------
    results : list, ModelfitResults
        List of ModelfitResults or single ModelfitResults
    include_all_estimation_steps : bool
        Whether to include all estimation steps, default is False

    Return
    ------
    pd.DataFrame
        A DataFrame of modelfit results with model name and estmation step as index.

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> from pharmpy.tools import summarize_modelfit_results
    >>> model = load_example_model("pheno")
    >>> summarize_modelfit_results(model.modelfit_results) # doctest: +SKIP
                     description  minimization_successful ...        ofv  ... runtime_total  ...
    pheno PHENOBARB SIMPLE MODEL                     True ... 586.276056  ...           4.0  ...
    """
    if isinstance(results, ModelfitResults):
        results = [results]

    if results is None:
        raise ValueError('Option `results` is None')
    if all(res is None for res in results):
        raise ValueError('All input results are empty')

    summaries = []

    for res in results:
        if res is not None:
            summary = _get_model_result_summary(res, include_all_estimation_steps)
            summary.insert(0, 'description', res.description)
            summaries.append(summary)

    with warnings.catch_warnings():
        # Needed because of warning in pandas 2.1.1
        warnings.filterwarnings(
            "ignore",
            message="The behavior of DataFrame concatenation with empty or all-NA entries is deprecated",
            category=FutureWarning,
        )

        df = pd.concat(summaries)

    return df


def _get_model_result_summary(res, include_all_estimation_steps=False):
    if not include_all_estimation_steps:
        summary_dict = _summarize_step(res, -1)
        index = pd.Index([res.name], name='model')
        summary_df = pd.DataFrame(summary_dict, index=index)
    else:
        summary_dicts = []
        tuples = []
        for i in range(len(res.evaluation)):
            summary_dict = _summarize_step(res, i)
            is_evaluation = res.evaluation.iloc[i]
            if is_evaluation:
                run_type = 'evaluation'
            else:
                run_type = 'estimation'
            summary_dict = {'run_type': run_type, **summary_dict}
            summary_dicts.append(summary_dict)
            tuples.append((res.name, i + 1))
        index = pd.MultiIndex.from_tuples(tuples, names=['model', 'step'])
        summary_df = pd.DataFrame(summary_dicts, index=index)

    no_of_errors = len(res.log.errors)
    no_of_warnings = len(res.log.warnings)

    minimization_idx = summary_df.columns.get_loc('minimization_successful')
    summary_df.insert(loc=minimization_idx + 1, column='errors_found', value=no_of_errors)
    summary_df.insert(loc=minimization_idx + 2, column='warnings_found', value=no_of_warnings)

    return summary_df


def _summarize_step(res, i):
    summary_dict = {}

    if i >= 0:
        minsucc = res.minimization_successful_iterations.iloc[i]
    else:
        minsucc = res.minimization_successful

    if minsucc is not None:
        summary_dict['minimization_successful'] = minsucc
    else:
        summary_dict['minimization_successful'] = False

    if i == -1 and res.ofv_iterations is not None:
        i = max(res.ofv_iterations.index.get_level_values(0)) - 1

    summary_dict['ofv'] = _get_ofv(res, i)
    summary_dict['runtime_total'] = res.runtime_total
    summary_dict['estimation_runtime'] = _get_estimation_runtime(res, i)

    pe = _get_parameter_estimates(res, i)
    ses = res.standard_errors
    rses = res.relative_standard_errors

    for param in pe.index:
        summary_dict[f'{param}_estimate'] = pe[param]
        if ses is not None:
            summary_dict[f'{param}_SE'] = ses[param]
        if rses is not None:
            summary_dict[f'{param}_RSE'] = rses[param]

    return summary_dict


def _get_ofv(res, i):
    if res.ofv_iterations is None:
        return res.ofv
    return res.ofv_iterations[i + 1,].iloc[-1]


def _get_parameter_estimates(res, i):
    if res.parameter_estimates_iterations is None:
        return res.parameter_estimates
    return res.parameter_estimates_iterations.loc[i + 1,].iloc[-1]


def _get_estimation_runtime(res, i):
    if res.estimation_runtime_iterations is None:
        return res.runtime_total
    return res.estimation_runtime_iterations.iloc[i]


def read_modelfit_results(path: Union[str, Path]) -> ModelfitResults:
    """Read results from external tool for a model

    Parameters
    ----------
    path : Path or str
        Path to model file

    Return
    ------
    ModelfitResults
        Results object
    """
    path = normalize_user_given_path(path)
    model = read_model(path)
    res = parse_modelfit_results(model, path)
    return res


def _get_run_setup_from_metadata(path):
    import pharmpy.workflows as workflows

    tool_database = workflows.default_tool_database(toolname=None, path=path, exist_ok=True)

    tool_metadata = tool_database.read_metadata()
    tool_name = tool_metadata['tool_name']
    common_options = tool_metadata['common_options']

    # TODO: Be more general
    dispatcher = getattr(workflows, common_options['dispatcher'].split('.')[-1])

    # TODO: Be more general
    assert common_options['database']['class'] == 'LocalDirectoryToolDatabase'
    assert common_options['database']['toolname'] == tool_name

    return dispatcher, tool_database


def load_example_modelfit_results(name: str):
    """Load the modelfit results of an example model

    Load the modelfit results of an example model built into Pharmpy

    Parameters
    ----------
    name : str
        Name of the model. Currently available models are "pheno" and "pheno_linear"

    Returns
    -------
    ModelfitResults
        Loaded modelfit results object

    Example
    -------
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> results = load_example_modelfit_results("pheno")
    >>> results.parameter_estimates
        PTVCL        0.004696
    PTVV         0.984258
    THETA_3      0.158920
    IVCL         0.029351
    IVV          0.027906
    SIGMA_1_1    0.013241
    Name: estimates, dtype: float64

    """
    available = ('moxo', 'pheno', 'pheno_linear')
    if name not in available:
        raise ValueError(f'Unknown example model {name}. Available examples: {available}')
    path = Path(__file__).resolve().parent.parent / 'modeling' / 'example_models' / (name + '.mod')
    res = read_modelfit_results(path)
    return res



"""
:meta private:
"""

import importlib
import inspect
import re
from functools import wraps
from typing import Callable

from .run import run_tool


def wrap(tool_name: str) -> Callable:
    tool_module = importlib.import_module(f'pharmpy.tools.{tool_name}')

    if 'create_workflow' not in dir(tool_module):
        raise ValueError(f'Tool `{tool_name}` does not export `create_workflow`.')

    create_workflow = tool_module.create_workflow

    if create_workflow.__doc__ is None:
        raise ValueError(
            f'Tool `{tool_name}` does export `create_workflow`'
            ' but this `create_workflow` is lacking `__doc__`.'
        )

    return _create_wrap(tool_name, create_workflow)


def _create_wrap(tool_name: str, create_workflow: Callable) -> Callable:
    @wraps(create_workflow)
    def func(*args, **kwargs):
        return run_tool(tool_name, *args, **kwargs)

    func.__name__ = f'run_{tool_name}'

    # NOTE: Override the signature to include **kwargs which are used in run_tool, needed for pharmr
    func.__signature__ = _append_kwargs_to_sig(inspect.signature(create_workflow))

    # NOTE: Add kwargs to docstring
    assert func.__doc__ is not None
    assert run_tool.__doc__ is not None
    func.__doc__ = _append_kwargs_to_doc(func.__doc__, run_tool.__doc__)

    return func


def _append_kwargs_to_sig(sig: inspect.Signature) -> inspect.Signature:
    param_kwargs = inspect.Parameter('kwargs', inspect.Parameter.VAR_KEYWORD)
    return sig.replace(parameters=tuple(sig.parameters.values()) + (param_kwargs,))


def _append_kwargs_to_doc(doc_wrapper: str, doc_run_tool: str) -> str:
    # NOTE: Get where in docstring to add kwargs documentation
    m_wrapper = re.compile(r'(.)\s*Returns*\s*\n\s*-------')
    search_wrapper = re.search(m_wrapper, doc_wrapper)
    assert search_wrapper is not None

    # NOTE: Get documentation for kwargs from run_tool
    m_run_tool = re.compile(r'(\s*kwargs\n(.+\n))\n.\s*Returns*\s*\n\s*-+')
    search_run_tool = re.search(m_run_tool, doc_run_tool)
    assert search_run_tool is not None

    return doc_wrapper.replace(search_wrapper.group(0), search_run_tool.group(0))



from functools import partial
from itertools import product
from typing import Iterable

from pharmpy.model import Model
from pharmpy.modeling import add_indirect_effect

from ..statement.feature.indirect_effect import IndirectEffect
from ..statement.feature.symbols import Name, Wildcard
from ..statement.statement import Statement
from .feature import Feature


def features(model: Model, statements: Iterable[Statement]) -> Iterable[Feature]:
    for statement in statements:
        if isinstance(statement, IndirectEffect):
            modes = (
                [Name('LINEAR'), Name('EMAX'), Name('SIGMOID')]
                if isinstance(statement.modes, Wildcard)
                else statement.modes
            )
            production = (
                [Name('PRODUCTION'), Name('DEGRADATION')]
                if isinstance(statement.production, Wildcard)
                else statement.production
            )

            params = list(product(modes, production))
            params = [(mode.name, production.name) for mode, production in params]

            for param in params:
                if param[1] == 'PRODUCTION':
                    yield ('INDIRECT', *param), partial(
                        add_indirect_effect, expr=param[0].lower(), prod=True
                    )
                elif param[1] == 'DEGRADATION':
                    yield ('INDIRECT', *param), partial(
                        add_indirect_effect, expr=param[0].lower(), prod=False
                    )



"""
:meta private:
"""
from typing import Literal, Optional

from pharmpy.deps import sympy
from pharmpy.model import (
    Assignment,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    Model,
    output,
)
from pharmpy.modeling import get_central_volume_and_clearance

from .error import set_proportional_error_model
from .expressions import _replace_trivial_redefinitions
from .odes import add_individual_parameter, set_first_order_elimination, set_initial_condition
from .parameter_variability import add_iiv

TMDD_TYPE = ('full', 'ib', 'cr', 'crib', 'qss', 'wagner', 'mmapp')
DV_TYPES = ('drug', 'drug_tot', 'target', 'target_tot', 'complex')


def set_tmdd(
    model: Model, type: Literal[TMDD_TYPE], dv_types: Optional[dict[Literal[DV_TYPES], int]] = None
):
    """Sets target mediated drug disposition

    Implemented target mediated drug disposition (TMDD) models are:

    - Full model
    - Irreversible binding approximation (IB)
    - Constant total receptor approximation (CR)
    - Irreversible binding and constant total receptor approximation (CR+IB)
    - Quasi steady-state approximation (QSS)
    - Wagner
    - Michaelis-Menten approximation (MMAPP)


    Parameters
    ----------
    model : Model
        Pharmpy model
    type : str
        Type of TMDD model
    dv_types: dict
        Dictionary of DV types for TMDD models with multiple DVs (e.g. dv_types = {'drug' : 1, 'target': 2}).
        For dv = 1 the only allowed keys are 'drug' and 'drug_tot'. If no DV for drug is specified then (free) drug
        will have dv = 1.

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_tmdd(model, "full")

    """
    if dv_types is not None:
        _validate_dv_types(dv_types)

    type = type.upper()

    model = _replace_trivial_redefinitions(model)
    model = set_first_order_elimination(model)

    odes = model.statements.ode_system
    central = odes.central_compartment
    cb = CompartmentalSystemBuilder(odes)

    vc, cl = get_central_volume_and_clearance(model)
    r_0 = sympy.Symbol('R_0')
    model = add_individual_parameter(model, r_0.name)
    model = add_iiv(model, [r_0], 'exp')
    kint = sympy.Symbol('KINT')
    model = add_individual_parameter(model, kint.name)

    y_symbol = _get_y_symbol(model)

    if type == "FULL":
        model, kon, koff, kdeg = _create_parameters(model, ['KON', 'KOFF', 'KDEG'])
        target_comp, complex_comp = _create_compartments(cb, ['TARGET', 'COMPLEX'])
        ksyn, ksyn_ass = _create_ksyn()

        cb.add_flow(target_comp, complex_comp, kon * central.amount / vc)
        cb.add_flow(complex_comp, target_comp, koff)
        cb.add_flow(target_comp, output, kdeg)
        cb.add_flow(complex_comp, output, kint)
        cb.set_input(target_comp, ksyn * vc)
        cb.set_input(
            central, koff * complex_comp.amount - kon * central.amount * target_comp.amount / vc
        )

        before = model.statements.before_odes + ksyn_ass
        after = model.statements.after_odes
    elif type == "IB":
        model, kdeg, kon = _create_parameters(model, ['KDEG', 'KON'])
        target_comp, complex_comp = _create_compartments(cb, ['TARGET', 'COMPLEX'])
        ksyn, ksyn_ass = _create_ksyn()

        cb.add_flow(target_comp, complex_comp, kon * central.amount / vc)
        cb.add_flow(target_comp, output, kdeg)
        cb.add_flow(complex_comp, output, kint)
        cb.set_input(target_comp, ksyn * vc)
        cb.set_input(central, -kon * central.amount * target_comp.amount / vc)

        before = model.statements.before_odes + ksyn_ass
        after = model.statements.after_odes
    elif type == "CR":
        model, kon, koff = _create_parameters(model, ['KON', 'KOFF'])
        complex_comp = _create_compartments(cb, ['COMPLEX'])

        cb.add_flow(complex_comp, central, koff + kon * central.amount / vc)
        cb.add_flow(complex_comp, output, kint)
        cb.add_flow(central, complex_comp, kon * r_0)

        before = model.statements.before_odes
        after = model.statements.after_odes
    elif type == "CRIB":
        model, kon = _create_parameters(model, ['KON'])
        complex_comp = _create_compartments(cb, ['COMPLEX'])

        cb.add_flow(complex_comp, output, kint)
        cb.add_flow(central, complex_comp, kon * r_0)
        cb.add_flow(complex_comp, central, kon * central.amount / vc)

        before = model.statements.before_odes
        after = model.statements.after_odes
    elif type == "QSS":
        model, kdc, kdeg = _create_parameters(model, ['KDC', 'KDEG'])
        target_comp = _create_compartments(cb, ['TARGET'])

        kd = sympy.Symbol('KD')

        ksyn, ksyn_ass = _create_ksyn()
        kd_ass = Assignment(kd, kdc * vc)

        lafree_symb = sympy.Symbol('LAFREE')
        lafree_expr = sympy.Rational(1, 2) * (
            central.amount
            - target_comp.amount
            - kd
            + sympy.sqrt((central.amount - target_comp.amount - kd) ** 2 + 4 * kd * central.amount)
        )
        lafree_ass = Assignment(lafree_symb, lafree_expr)

        num_peripheral_comp = len(odes.find_peripheral_compartments())
        if num_peripheral_comp == 0:
            elimination_rate = odes.get_flow(central, output)
            cb.remove_flow(central, output)
            cb.set_input(
                central,
                -lafree_symb * elimination_rate
                - target_comp.amount * kint * lafree_symb / (kd + lafree_symb),
            )
            cb.set_input(
                target_comp,
                ksyn * vc
                - kdeg * target_comp.amount
                - (kint - kdeg) * target_comp.amount * lafree_symb / (kd + lafree_symb),
            )

            # FIXME: Should others also have flows?
            central = cb.find_compartment('CENTRAL')
            cb.add_flow(central, output, lafree_symb * elimination_rate)

            lafreef = sympy.Symbol("LAFREEF")
            lafree_final = Assignment(lafreef, lafree_expr)
            before = model.statements.before_odes + (ksyn_ass, kd_ass, lafree_ass)
            after = lafree_final + model.statements.after_odes
            ipred = lafreef / vc
            after = after.reassign(y_symbol, ipred)
        elif num_peripheral_comp > 0 and num_peripheral_comp <= 2:
            peripheral1 = _create_compartments(cb, ['PERIPHERAL1'])
            flow_central_peripheral1 = odes.get_flow(central, peripheral1)
            if num_peripheral_comp == 2:
                peripheral2 = _create_compartments(cb, ['PERIPHERAL2'])
                flow_central_peripheral2 = odes.get_flow(central, peripheral2)

            elimination_rate = odes.get_flow(central, output)
            cb.remove_flow(central, output)
            if num_peripheral_comp == 1:
                cb.set_input(
                    central,
                    -target_comp.amount * kint * lafree_symb / (kd + lafree_symb)
                    - lafree_symb * flow_central_peripheral1
                    + flow_central_peripheral1 * central.amount,
                )
                cb.set_input(
                    target_comp,
                    ksyn * vc
                    - kdeg * target_comp.amount
                    - (kint - kdeg) * target_comp.amount * lafree_symb / (kd + lafree_symb),
                )
                cb.set_input(
                    peripheral1,
                    lafree_symb * flow_central_peripheral1
                    - flow_central_peripheral1 * central.amount,
                )
            elif num_peripheral_comp == 2:
                cb.set_input(
                    central,
                    -target_comp.amount * kint * lafree_symb / (kd + lafree_symb)
                    - lafree_symb * flow_central_peripheral1
                    - lafree_symb * flow_central_peripheral2
                    + flow_central_peripheral2 * central.amount,
                )
                cb.set_input(
                    target_comp,
                    ksyn * vc
                    - kdeg * target_comp.amount
                    - (kint - kdeg) * target_comp.amount * lafree_symb / (kd + lafree_symb),
                )
                cb.set_input(
                    peripheral1,
                    lafree_symb * flow_central_peripheral1
                    - flow_central_peripheral1 * central.amount,
                )
                cb.set_input(
                    peripheral2,
                    lafree_symb * flow_central_peripheral2
                    - flow_central_peripheral2 * central.amount,
                )

            central = cb.find_compartment('CENTRAL')
            cb.add_flow(central, output, lafree_symb * elimination_rate / central.amount)

            lafreef = sympy.Symbol("LAFREEF")
            lafree_final = Assignment(lafreef, lafree_expr)
            before = model.statements.before_odes + (ksyn_ass, kd_ass, lafree_ass)
            after = lafree_final + model.statements.after_odes
            ipred = lafreef / vc
            after = after.reassign(y_symbol, ipred)
        else:
            raise ValueError('More than 2 peripheral compartments are not supported.')
    elif type == 'WAGNER':
        model, km = _create_parameters(model, ['KM'])

        kel = odes.get_flow(central, output)
        kd = km * vc
        rinit = r_0 * vc
        rinit_ass = Assignment(sympy.Symbol('RINIT'), rinit)
        kd_ass = Assignment(sympy.Symbol('KD'), km * vc)

        lafree_symb = sympy.Symbol('LAFREE')
        lafree_expr = sympy.Rational(1, 2) * (
            central.amount
            - rinit
            - kd
            + sympy.sqrt((central.amount - rinit - kd) ** 2 + 4 * kd * central.amount)
        )
        lafree_ass = Assignment(lafree_symb, lafree_expr)

        num_peripheral_comp = len(odes.find_peripheral_compartments())
        if num_peripheral_comp == 0:
            cb.add_flow(central, output, kel)
            cb.set_input(
                central,
                kint * lafree_symb
                - kint * central.amount
                - kel * lafree_symb
                + kel * central.amount,
            )
        elif num_peripheral_comp == 1:
            peripheral = _create_compartments(cb, ['PERIPHERAL1'])
            kcp = odes.get_flow(central, peripheral)
            cb.add_flow(central, output, kel)
            cb.set_input(
                central,
                kint * lafree_symb
                - kint * central.amount
                - kel * lafree_symb
                + kel * central.amount
                - kcp * lafree_symb
                + kcp * central.amount,
            )
            cb.set_input(peripheral, kcp * lafree_symb - kcp * central.amount)
        elif num_peripheral_comp == 2:
            peripheral1 = _create_compartments(cb, ['PERIPHERAL1'])
            kcp1 = odes.get_flow(central, peripheral1)
            peripheral2 = _create_compartments(cb, ['PERIPHERAL2'])
            kcp2 = odes.get_flow(central, peripheral2)

            cb.add_flow(central, output, kel)
            cb.set_input(
                central,
                kint * lafree_symb
                - kint * central.amount
                - kel * lafree_symb
                + kel * central.amount
                - kcp1 * lafree_symb
                - kcp2 * lafree_symb,
            )
            cb.set_input(peripheral1, kcp1 * lafree_symb - kcp1 * central.amount)
            cb.set_input(peripheral2, kcp2 * lafree_symb - kcp2 * central.amount)

        lafreef = sympy.Symbol("LAFREEF")
        lafree_final = Assignment(lafreef, lafree_expr)
        before = model.statements.before_odes + lafree_ass + kd_ass + rinit_ass
        after = lafree_final + model.statements.after_odes
        ipred = lafreef / vc
        after = after.reassign(y_symbol, ipred)
    elif type == 'MMAPP':
        model, km, kdeg = _create_parameters(model, ['KM', 'KDEG'])
        target_comp = _create_compartments(cb, ['TARGET'])
        ksyn, ksyn_ass = _create_ksyn()

        target_elim = kdeg + (kint - kdeg) * central.amount / vc / (km + central.amount / vc)
        cb.add_flow(target_comp, output, target_elim)
        elim = cl / vc
        cb.add_flow(central, output, elim)
        cb.set_input(target_comp, ksyn)
        cb.set_input(
            central, -target_comp.amount * central.amount * kint / (central.amount / vc + km)
        )

        before = model.statements.before_odes + ksyn_ass
        after = model.statements.after_odes
    else:
        raise ValueError(f'Unknown TMDD type "{type}".')

    model = model.replace(statements=before + CompartmentalSystem(cb) + after)
    if type not in ['CR', 'CRIB', 'WAGNER', 'MMAPP']:
        model = set_initial_condition(model, "TARGET", r_0 * vc)
    if type == 'MMAPP':
        model = set_initial_condition(model, "TARGET", r_0)

    # Multiple DVs:
    if dv_types is not None:
        if type in ['FULL', 'IB']:
            if 'drug_tot' in dv_types.keys():
                new_y = (central.amount + complex_comp.amount) / vc
                after = model.statements.after_odes
                after = after.reassign(y_symbol, new_y)
                model = model.replace(
                    statements=model.statements.before_odes + model.statements.ode_system + after
                )
            if 'target' in dv_types.keys():
                y_target = sympy.Symbol("Y_TARGET")
                ytarget = Assignment(y_target, target_comp.amount / vc)
                dvs = model.dependent_variables.replace(y_target, dv_types['target'])
                model = model.replace(
                    statements=model.statements + ytarget, dependent_variables=dvs
                )
            if 'complex' in dv_types.keys():
                y_complex = sympy.Symbol("Y_COMPLEX")
                ycomplex = Assignment(y_complex, complex_comp.amount / vc)
                dvs = model.dependent_variables.replace(y_complex, dv_types['complex'])
                model = model.replace(
                    statements=model.statements + ycomplex, dependent_variables=dvs
                )
            if 'target_tot' in dv_types.keys():
                y_target_tot = sympy.Symbol("Y_TOTTARGET")
                ytargettot = Assignment(
                    y_target_tot, (target_comp.amount + complex_comp.amount) / vc
                )
                dvs = model.dependent_variables.replace(y_target_tot, dv_types['target_tot'])
                model = model.replace(
                    statements=model.statements + ytargettot, dependent_variables=dvs
                )
        elif type == 'QSS':
            if 'drug_tot' in dv_types.keys():
                new_y = central.amount / vc
                after = model.statements.after_odes
                after = after.reassign(y_symbol, new_y)
                model = model.replace(
                    statements=model.statements.before_odes + model.statements.ode_system + after
                )
            if 'target' in dv_types.keys():
                y_target = sympy.Symbol("Y_TARGET")
                ytarget = Assignment(y_target, (target_comp.amount - central.amount + lafreef) / vc)
                dvs = model.dependent_variables.replace(y_target, dv_types['target'])
                model = model.replace(
                    statements=model.statements + ytarget, dependent_variables=dvs
                )
            if 'complex' in dv_types.keys():
                y_complex = sympy.Symbol("Y_COMPLEX")
                ycomplex = Assignment(y_complex, (central.amount - lafreef) / vc)
                dvs = model.dependent_variables.replace(y_complex, dv_types['complex'])
                model = model.replace(
                    statements=model.statements + ycomplex, dependent_variables=dvs
                )
            if 'target_tot' in dv_types.keys():
                y_target_tot = sympy.Symbol("Y_TOTTARGET")
                ytargettot = Assignment(y_target_tot, target_comp.amount / vc)
                dvs = model.dependent_variables.replace(y_target_tot, dv_types['target_tot'])
                model = model.replace(
                    statements=model.statements + ytargettot, dependent_variables=dvs
                )
        elif type == 'MMAPP':
            if 'target' in dv_types.keys():
                y_target = sympy.Symbol("Y_TARGET")
                ytarget = Assignment(y_target, target_comp.amount / vc)
                dvs = model.dependent_variables.replace(y_target, dv_types['target'])
                model = model.replace(
                    statements=model.statements + ytarget, dependent_variables=dvs
                )
            if 'target_tot' in dv_types.keys():
                y_target_tot = sympy.Symbol("Y_TOTTARGET")
                ytargettot = Assignment(y_target_tot, target_comp.amount / vc)
                dvs = model.dependent_variables.replace(y_target_tot, dv_types['target_tot'])
                model = model.replace(
                    statements=model.statements + ytargettot, dependent_variables=dvs
                )
        elif type in ['CR', 'CRIB']:
            if 'drug_tot' in dv_types.keys():
                new_y = (central.amount + complex_comp.amount) / vc
                after = model.statements.after_odes
                after = after.reassign(y_symbol, new_y)
                model = model.replace(
                    statements=model.statements.before_odes + model.statements.ode_system + after
                )
            if 'complex' in dv_types.keys():
                y_complex = sympy.Symbol("Y_COMPLEX")
                ycomplex = Assignment(y_complex, complex_comp.amount / vc)
                dvs = model.dependent_variables.replace(y_complex, dv_types['complex'])
                model = model.replace(
                    statements=model.statements + ycomplex, dependent_variables=dvs
                )
        elif type == 'WAGNER':
            if 'drug_tot' in dv_types.keys():
                new_y = central.amount / vc
                after = model.statements.after_odes
                after = after.reassign(y_symbol, new_y)
                model = model.replace(
                    statements=model.statements.before_odes + model.statements.ode_system + after
                )
            if 'complex' in dv_types.keys():
                y_complex = sympy.Symbol("Y_COMPLEX")
                ycomplex = Assignment(y_complex, (central.amount - lafreef) / vc)
                dvs = model.dependent_variables.replace(y_complex, dv_types['complex'])
                model = model.replace(
                    statements=model.statements + ycomplex, dependent_variables=dvs
                )

        # Add proportional error model
        if sympy.Symbol('Y_TARGET') in list(model.dependent_variables):
            model = set_proportional_error_model(model, dv=dv_types['target'])
        if sympy.Symbol('Y_COMPLEX') in list(model.dependent_variables):
            model = set_proportional_error_model(model, dv=dv_types['complex'])
        if sympy.Symbol('Y') in list(model.dependent_variables) and 'drug_tot' in dv_types.keys():
            model = set_proportional_error_model(model, dv=dv_types['drug_tot'])
        if sympy.Symbol('Y_TOTTARGET') in list(model.dependent_variables):
            model = set_proportional_error_model(model, dv=dv_types['target_tot'])

    return model.update_source()


def _create_parameters(model, names):
    symbs = []
    for name in names:
        symb = sympy.Symbol(name)
        symbs.append(symb)
        model = add_individual_parameter(model, symb.name)
    return model, *symbs


def _create_compartments(cb, names):
    comps = []
    for name in names:
        comp = Compartment.create(name=name)
        comps.append(comp)
        cb.add_compartment(comp)
    if len(comps) == 1:
        return comps[0]
    else:
        return comps


def _create_ksyn():
    ksyn = sympy.Symbol('KSYN')
    ksyn_ass = Assignment(ksyn, sympy.Symbol("R_0") * sympy.Symbol("KDEG"))
    return ksyn, ksyn_ass


def _validate_dv_types(dv_types):
    # Make sure that values are unique
    assert len(dv_types.values()) == len(set(dv_types.values()))
    # Validate keys
    for key, value in dv_types.items():
        if key not in ['drug', 'target', 'complex', 'drug_tot', 'target_tot']:
            raise ValueError(
                f'Invalid dv_types key "{key}". Allowed keys are:'
                f' "drug", "target", "complex", "drug_tot" and "target_tot".'
            )
        if key not in ['drug', 'drug_tot'] and value == 1:
            raise ValueError('Only drug can have DVID = 1. Please choose another DVID.')


def _get_y_symbol(model):
    t = sympy.Symbol("t")
    y_statement = model.statements.find_assignment("Y")
    if t in y_statement.free_symbols:
        return y_statement.symbol
    else:
        sset = model.statements.direct_dependencies(y_statement)
        for s in sset:
            if sympy.Symbol("t") in s.free_symbols:
                return s.symbol
